/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaMath.h
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Supplies Cuda math functions for pitches.
**
*****************************************************************************/

#pragma once

#ifndef __SVL_CUDA_MATH_H
#define __SVL_CUDA_MATH_H

#include <iostream>
#include <vector>

#include "cv.h"
#include "cxcore.h"

#include "svlCudaCommon.h"
#include "svlCudaMemHandler.h"
using namespace std;

class svlCudaMath {
public:
	svlCudaMath();
	virtual ~svlCudaMath();

	// *** Elementary math operations.
	// Subtract one pitch from another: <a -= b;>.
	inline static void subtract(svlCudaPitch *a, svlCudaPitch *b, int img_w = -1, int img_h = -1) {
		svlCudaSubtract(*a, *b, img_w, img_h);
	}
	// Subsample a in x and y, save to b.
	inline static void decimate(svlCudaPitch *a, svlCudaPitch *b, int delta_x = 2, int delta_y = 2, int img_w = -1, int img_h = -1) {
		svlCudaDecimate(*a, *b, delta_x, delta_y, img_w, img_h);
	}
	// Add in quadrature: a <- sqrt( a.^2 + b.^2 );
	inline static void addQuadrature(svlCudaPitch *a, svlCudaPitch *b, int img_w = -1, int img_h = -1) {
		svlCudaAddQuadrature(a, b, img_w, img_h);
	}

	// *** Min-max operations.
	// Do max operation over variable width and height (only correct for width and height <= 16).
	inline static void maxWindow(svlCudaPitch *pitch_in, svlCudaPitch *pitch_out,
		int max_w = 7, int max_h = 7, int img_w = -1, int img_h = -1, svlCudaCoveringGrid *cudagrid = NULL) {
		svlCudaMax(*pitch_in, *pitch_out, max_w, max_h, img_w, img_h, cudagrid);
	}
	// Compute max in all 7-by-7 windows of pitch_in.
	inline static void max7x7(svlCudaPitch *pitch_in, svlCudaPitch *pitch_out, int img_w = -1, int img_h = -1, svlCudaCoveringGrid *cudagrid = NULL) {
		svlCudaMax7x7(*pitch_in, *pitch_out, img_w, img_h, cudagrid);
	}
	// Compute max in all 7-by-7 (centered) windows of pitch_in.
	inline static void max7x7Cent(svlCudaPitch *pitch_in, svlCudaPitch *pitch_out, int img_w = -1, int img_h = -1) {
		svlCudaMax7x7Cent(*pitch_in, *pitch_out, img_w, img_h);
	}
	// Compute min and max in all 4-by-4 windows of pitch_in.
	inline static void minMax4x4(svlCudaPitch *pitch_in, svlCudaPitch *min_out, svlCudaPitch *max_out, int img_w = -1, int img_h = -1) {
		svlCudaMinMax4x4(*pitch_in, *min_out, *max_out, img_w, img_h);
	}
	// Calculate min and max over 32-by-32 windows.
	inline static void minMax32x32(svlCudaPitch *pitch_in, svlCudaPitch *min_out, svlCudaPitch *max_out, int img_w = -1, int img_h = -1) {
		svlCudaMinMax32x32(*pitch_in, *min_out, *max_out, img_w, img_h);
	}

	// *** Miscellaneous operations
	// Integral and integral-squared images.
	inline static void integralLine(svlCudaLine *n, svlCudaLine *integ, int w = -1, int dim = 128, bool do_cumsum = false, int stride = -1) {
		svlCudaIntegralLine(n, integ, w, dim, do_cumsum, stride);
	}
	// Warning: deactivated because it apparently does not work. "integralBlock" does, though, and is slightly faster.
	/*inline static void integral(svlCudaPitch *pin, svlCudaPitch *integ, svlCudaPitch *integ2 = NULL, int img_w = -1, int img_h = -1, int dim = 128) {
		svlCudaIntegral(*pin, integ, integ2, img_w, img_h, dim);
	}*/
	inline static void integralBlock(svlCudaPitch *pin, svlCudaPitch *integ, svlCudaPitch *tempint, svlCudaPitch *integ2 = NULL, svlCudaPitch *tempint2 = NULL, int img_w = -1, int img_h = -1, int dim = 128) {
		svlCudaIntegralBlock(*pin, integ, tempint, integ2, tempint2, img_w, img_h, dim);
	}
	inline static void rowIntegralBlock(svlCudaPitch *pin, svlCudaPitch *integ, svlCudaPitch *integ2 = NULL, int img_w = -1, int img_h = -1, int dim = 128) {
		svlCudaRowIntegralBlock(*pin, integ, integ2, img_w, img_h, dim);
	}

	// Image resizing.
	inline static void resize2D(svlCudaPitch *pin, svlCudaPitch *pout, float scale = 0.8f, int img_w = -1, int img_h = -1) {
		svlCudaResize2D(*pin, *pout, scale, img_w, img_h);
	}

	// Round a/b to nearest higher integer value
	static int iDivUp(int a, int b) {
		return (a % b != 0) ? (a / b + 1) : (a / b);
	}
	// Round a/b to nearest lower integer value
	static int iDivDown(int a, int b) {
		return a / b;
	}
	// Align a to nearest higher multiple of b
	static int iAlignUp(int a, int b) {
		return (a % b != 0) ?  (a - a % b + b) : a;
	}
	// Align a to nearest lower multiple of b
	static int iAlignDown(int a, int b) {
		return a - a % b;
	}
	static inline int iAlignUpPow2(int a) {
		return int( pow( 2.0, ceil( log((double)a) / log(2.0) ) ) );
	}
};


#endif

