/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaMemHandler.h
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Handles NVIDIA Cuda memory interface.
**
*****************************************************************************/

#pragma once

#ifndef __SVL_CUDA_MEM_HANDLER
#define __SVL_CUDA_MEM_HANDLER

#include <iostream>
#include <vector>
#include <list>

#include "cv.h"
#include "cxcore.h"

#include "svlCudaCommon.h"
using namespace std;

typedef pair<svlCudaLine*,char*> NamedLine;
typedef pair<svlCudaPitch*,char*> NamedPitch;

class svlCudaMemHandler {
protected:
	static list<NamedLine> nls;
	static list<NamedPitch> nps;
	static bool track_allocations;

public:
	svlCudaMemHandler();
	virtual ~svlCudaMemHandler();

	// Memory allocation tracking functions.
	static bool getTracking() { return track_allocations; }
	static void setTracking(bool in_the_track) { track_allocations = in_the_track; }
	static void printTotalAllocation();
	static void printAllAllocations();

	static bool testMemoryAccess(int w = 640, int h = 480);

	// *** Memory management functions ***
	// Allocate device memory in the form a pitch and return addressable index.
	static svlCudaPitch* allocPitch(int w, int h, char *name = NULL) {
		svlCudaPitch *ret = new svlCudaPitch();
		svlCudaAllocPitch(w, h, *ret);
		if ( track_allocations )
			nps.push_back( NamedPitch(ret, name) );
		return ret;
	}
	// Allocate device memory in the form of pitches.
	static inline void allocPitches(int w, int h, vector<svlCudaPitch*> &pitches, int num = 1) {
		svlCudaAllocPitches(w, h, pitches, num);
	}
	static void freePitch(svlCudaPitch *p) {
		if ( track_allocations ) {
			// Remove from list of allocations.
			for ( list<NamedPitch>::iterator it = nps.begin(); it != nps.end(); ++it ) {
				if ( it->first == p ) {
					svlCudaFreePitch(*p);
					nps.erase(it);
					return;
				}
			}
			printf("ERROR: Freeing un-allocated device memory: %08x\n", p->data);
		}
		else
			svlCudaFreePitch(*p); // Just free it.
	}
	// Allocate linear device memory.
	static svlCudaLine* allocLine(int w, char *name = NULL) {
		svlCudaLine *ret = new svlCudaLine();
		svlCudaAllocLine(w, *ret);
		if ( track_allocations )
			nls.push_back( NamedLine(ret, name) );
		return ret;
	}
	static void freeLine(svlCudaLine *n) {
		if ( track_allocations ) {
			// Remove from list of allocations.
			for ( list<NamedLine>::iterator it = nls.begin(); it != nls.end(); ++it ) {
				if ( it->first == n ) {
					svlCudaFreeLine(*n);
					nls.erase(it);
					return;
				}
			}
			printf("ERROR: Freeing un-allocated device memory: %x\n", n->data);
		}
		else
			svlCudaFreeLine(*n); // Just free it.
	}

	// *** Memory transfer functions ***
	// Line functions.
	// Copy from GPU to linear memory.
	static inline void lineFromGPU(float *dest, svlCudaLine *n, int w = -1) {
		svlCudaLineFromGPU(dest, *n, w);
	}
	// Copy linear memory to GPU pitch.
	static inline void lineToGPU(svlCudaLine *n, float *src, int w = -1) {
		svlCudaLineToGPU(*n, src, w);
		n->source = (void*)src;
	}
	// Pitch functions.
	// Copy from GPU to linear memory.
	static inline void pitchFromGPU(float *dest, svlCudaPitch *p) {
		svlCudaPitchFromGPU(dest, *p);
	}
	// Copy linear memory to GPU pitch.
	static inline void pitchToGPU(svlCudaPitch *p, float *src) {
		svlCudaPitchToGPU(*p, src);
		p->source = (void*)src;
	}
	static inline void pitchToPitch(svlCudaPitch *dest, svlCudaPitch *src, int img_w = -1, int img_h = -1) {
		svlCudaPitchToPitch(*dest, *src, img_w, img_h);
		dest->source = src->source;
	}
	// Set memory.
	static inline void pitchSet(svlCudaPitch *p, int value = 0) {
		svlCudaMemset2D(*p, value);
	}
	// Copy linear float to a pitch on the GPU.
	static inline void floatToGPU(float *src, svlCudaPitch *p) {
		svlCudaPitchToGPU(*p, src);
		p->source = (void*)src;
	}
	// Copy float Ipl to GPU (TODO paulb: this as an optimized template).
	static inline void floatIplToGPU(const IplImage *im, svlCudaPitch *p) {
		svlCudaPitchToGPU_all(*p, (float*)im->imageData, im->widthStep, im->width, im->height);
	}
	// Copy a pitch out from the GPU into dest.
	static inline void floatFromGPU(float *dest, svlCudaPitch *p) {
		svlCudaPitchFromGPU(dest, *p);
	}
	// Copy a pitch out from the GPU into dest using width and height.
	static inline void floatFromGPU(float *dest, int img_w, int img_h, svlCudaPitch *p) {
		svlCudaPitchFromGPU(dest, *p, img_w, img_h);
	}
	// Sending IplImages to/from the GPU.
	// Copy a 1-channel, float IplImage out of a pitch from index ix.
	static inline void floatIplFromGPU(IplImage *im, svlCudaPitch *p) {
		svlCudaPitchFromGPU_all((float*)im->imageData, im->widthStep, im->width, im->height, *p);
	}
	// Templates.	
	template <class C> static void IplToGPU(const IplImage *im, svlCudaPitch *p, int row_offset = 0, int col_offset = 0);
	template <class C> static void IplFromGPU(IplImage *im, svlCudaPitch *p);
	template <class C> static void IplToGPUasync(const IplImage *im, svlCudaPitch *p, int row_offset = 0, int col_offset = 0);

	// Random-access indexing functions.
	static int fillIndexImage(const vector<CvPoint> &locations, IplImage *ixage); // Returns max populated row.
	static void fillIndexLine(const vector<CvPoint> &locations, svlCudaLine *ixline);
	inline static void addCvPoint2Line(svlCudaLine *lin, CvPoint &pt, svlCudaLine *lout, int w = -1) {
		svlCudaAddCvPoint2Line(lin, pt, lout, w);
	}
	// Row random access functions.
	static inline void rowRandomAccess(const IplImage *ixes, IplImage *data_out, const vector<int> row_counts, float *dest, svlCudaPitch *data_p, svlCudaPitch *addy_p) {
		svlCudaRowRandomAccess(ixes, data_out, row_counts, dest, *data_p, *addy_p);
	}
	// Linear stacked random access.
	static inline void linearRandomAccess(const IplImage *ixes, int max_row, float *dest, svlCudaPitch *data_p, svlCudaPitch *addy_p, svlCudaPitch *data_out_p) {
		svlCudaLinearRandomAccess(ixes, max_row, dest, *data_p, *addy_p, *data_out_p);
	}
	// Linear stacked line-based random access.
	static inline void linear2DRandomAccess(svlCudaLine *ixline, svlCudaPitch *data, svlCudaLine *res, int res_off, int res_stride, int res_n) {
		svlCudaLine2DRandomAccess(ixline, data, res, res_off, res_stride, res_n);
	}
	template<class C_IN> static void vector2FloatLine(const vector<C_IN> &vec, svlCudaLine *line);
	template<class C_IN> static void vector2IntLine(const vector<C_IN> &vec, svlCudaLine *line);
	static void CvPoint2IntLine(const vector<CvPoint> &vec, svlCudaLine *line, bool yx = false);

	// File I/O.
	static void pitch2file(const char *file, svlCudaPitch *p);
	static void line2file(const char *file, svlCudaLine *n);

	// *** Miscellaneous access functions ***
	static inline bool pitchHasCached(void *ptr, svlCudaPitch *p) {
		return p->source == ptr;
	}
};

// Copy a 1-channel, general-type IplImage over to the GPU to the pitch at index "ix."
template <class C>
void svlCudaMemHandler::IplToGPU(const IplImage *im, svlCudaPitch *p, int row_offset, int col_offset)
{
	// TODO Paul B: Option to do pre-zeroed full image padding.
	/*assert( ix < int(pitches.size()) );
	assert( im->width < pitches[ix].w && im->height < pitches[ix].h );*/

	float *data = (float*)malloc( sizeof(float)*( im->width - col_offset )*( im->height - row_offset ) );
	float *ptr = data;
	for ( int i=row_offset; i<im->height; ++i ) {
		for ( int j=col_offset; j<im->width; ++j ) {
			*ptr++ = (float)CV_IMAGE_ELEM(im, C, i, j);
		}
	}
	svlCudaPitchToGPU(*p, data, im->width - col_offset, im->height - row_offset);
	p->source = (void*)im;
	//printf("cached %d as %x\n", ix, (void*)im);
	free(data);
}
// Above optimized for float.
//template <>
//void svlCudaMemHandler::IplToGPU<float>(const IplImage *im, int ix, int row_offset, int col_offset)
//{
//	svlCudaPitchFromGPU_all((float*)im->imageData, im->widthStep, im->width, im->height, pitches[ix]);
//	pitches[ix].source = (void*)im;
//}

// Copy a 1-channel, general-type IplImage over to the GPU to the pitch at index "ix."
template <class C>
void svlCudaMemHandler::IplToGPUasync(const IplImage *im, svlCudaPitch *p, int row_offset, int col_offset)
{
	// TODO Paul B: Option to do pre-zeroed full image padding.
	//assert( im->width < pitches[ix].w && im->height < pitches[ix].h );

	float *data = (float*)malloc( sizeof(float)*( im->width - col_offset )*( im->height - row_offset ) );
	float *ptr = data;
	for ( int i=row_offset; i<im->height; ++i ) {
		for ( int j=col_offset; j<im->width; ++j ) {
			*ptr++ = (float)CV_IMAGE_ELEM(im, C, i, j);
		}
	}
	svlCudaPitchToGPUasync(*p, data, im->width - col_offset, im->height - row_offset);
	p->source = (void*)im;
	free(data);
}

// Copy a 1-channel, general-type IplImage out of a pitch from index "ix."
template <class C>
void svlCudaMemHandler::IplFromGPU(IplImage *im, svlCudaPitch *p)
{
	// TODO Paul B: Option to do pre-zeroed full image padding?
	//assert( im->width < pitches[ix].w && im->height < pitches[ix].h );

	float *data = (float*)malloc(sizeof(float)*im->width*im->height);
	svlCudaPitchFromGPU(data, *p, im->width, im->height);
	float *ptr = data;
	for ( int i=0; i<im->height; ++i ) {
		for ( int j=0; j<im->width; ++j ) {
			CV_IMAGE_ELEM(im, C, i, j) = (C)*ptr;
			++ptr;
		}
	}
	free(data);
}

// Fill a line with the contents of a vector.
template<class C_IN>
void svlCudaMemHandler::vector2FloatLine(const vector<C_IN> &vec, svlCudaLine *line)
{
	// Collaps row-column indices into line linearly.
	float *data = (float*)malloc(sizeof(float)*int(vec.size()));
	float *ptr = data;
	for ( vector<C_IN>::const_iterator v=vec.begin(); v!=vec.end(); ++v ) {
		*ptr++ = (float)*v;
	}
	lineToGPU(line, data, int(vec.size()));
	free(data);
}

// Fill an int line with the contents of a vector.
template<class C_IN>
void svlCudaMemHandler::vector2IntLine(const vector<C_IN> &vec, svlCudaLine *line)
{
	// Collaps row-column indices into line linearly.
	int *data = (int*)malloc(sizeof(int)*int(vec.size()));
	int *ptr = data;
	for ( vector<C_IN>::const_iterator v=vec.begin(); v!=vec.end(); ++v ) {
		*ptr++ = (int)*v;
	}
	lineToGPU(line, (float*)data, int(vec.size()));
	free(data);
}

#endif

