/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlBoostedClassifierSet.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**
*****************************************************************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <limits>
#include <vector>

#include "cv.h"
#include "cxcore.h"
#include "ml.h"

#include "svlBase.h"
#include "svlBoostedClassifierSet.h"

using namespace std;

#define MULTITHREAD_EVAL
#ifdef MULTITHREAD_EVAL
struct svlBoostedClassifierEvalArgs {
    svlBoostedClassifierEvalArgs(const vector<vector<double> > *x,
        vector<svlSmartPointer<CvBoost> > *m, vector<vector<double> > *y,
        int i) : _x(x), _models(m), _y(y), _indx(i) {
        // do nothing
    }

    const vector<vector<double> > * const _x;           // feature vectors
    vector<svlSmartPointer<CvBoost> > * const _models;  // boosted classifiers
    vector<vector<double> > * const _y;                 // return values
    const int _indx;                                    // model index
};

void *fcnEvalBoostedClassifier(void *a, unsigned tid)
{
    svlBoostedClassifierEvalArgs *args = (svlBoostedClassifierEvalArgs *)a;

#if 0
    int length = cvSliceLength(CV_WHOLE_SEQ, 
        args->_models->at(args->_indx)->get_weak_predictors());
    CvMat *weakResponses = cvCreateMat(length, 1, CV_32FC1);
    CvMat *fv = cvCreateMat(1, (int)args->_x->at(0).size(), CV_32FC1);

    for (int j = 0; j < (int)args->_x->size(); j++) {
        for (int k = 0; k < fv->width; k++) {
            CV_MAT_ELEM(*fv, float, 0, k) = (float)args->_x->at(j)[k];
        }

        args->_models->at(args->_indx)->predict(fv, NULL, weakResponses, CV_WHOLE_SEQ);
        args->_y->at(j)[args->_indx] = cvSum(weakResponses).val[0];
    }

    cvReleaseMat(&fv);
    cvReleaseMat(&weakResponses);
#else
    // Steve's OpenCV hack to speed things up
    CvBoost *model = args->_models->at(args->_indx); 
    CvSeq *weak = model->get_weak_predictors();

    CvSeqReader reader;
    cvStartReadSeq(weak, &reader);
    
    for (int j = 0; j < (int)args->_x->size(); j++) {
        cvSetSeqReaderPos(&reader, 0);
    
        double y = 0;
        for (int i = 0; i < weak->total; i++) {
            CvBoostTree* wtree;
            CV_READ_SEQ_ELEM(wtree, reader);  

            const CvDTreeNode* node = wtree->get_root();
            while (node->left) {
                CvDTreeSplit* split = node->split;
                int vi = split->var_idx;
                int dir = args->_x->at(j)[vi] <= split->ord.c ? -1 : 1;
                if (split->inversed)
                    dir = -dir;           
                node = (dir < 0) ? node->left : node->right;
            }
            
            y += node->value;
        }

        args->_y->at(j)[args->_indx] = y;
    }
#endif
    
    delete args;
    return NULL;
}
#endif

// svlBoostedClassifierSet class ---------------------------------------------

svlBoostedClassifierSet::svlBoostedClassifierSet(unsigned n) : svlOptions()
{
    _models.resize(n);
    
    // default training options
    declareOption("boostingRounds", "100");
    declareOption("trimRate", "0.9");
    declareOption("numSplits", "2");
    declareOption("pseudoCounts", "1");
}

svlBoostedClassifierSet::svlBoostedClassifierSet(const svlBoostedClassifierSet& bcs) :
    svlOptions(bcs), _models(bcs._models)
{
    // do nothing
}

svlBoostedClassifierSet::~svlBoostedClassifierSet()
{
    // do nothing
}

// i/o
void svlBoostedClassifierSet::initialize(unsigned n)
{
    _models.clear();
    _models.resize(n);
}

bool svlBoostedClassifierSet::save(const char *filestem, const char *ext, int modelIndx)
{
    for (int i = 0; i < (int)_models.size(); i++) {
        string modelFilename = string(filestem) + string(".") +
            toString(i) + string(ext);
        if ((modelIndx < 0) || (i == modelIndx)) {
            SVL_LOG(SVL_LOG_VERBOSE, "...svlBoostedClassifierSet saving to " << modelFilename);
            _models[i]->save(modelFilename.c_str());
        } else {
            SVL_LOG(SVL_LOG_VERBOSE, "...svlBoostedClassifierSet skipping to " << modelFilename);
        }
    }

    return true;
}

bool svlBoostedClassifierSet::load(const char *filestem, const char *ext, int modelIndx)
{
    for (int i = 0; i < (int)_models.size(); i++) {
        string modelFilename = string(filestem) + string(".") +
            toString(i) + string(ext);
        if ((modelIndx < 0) || (i == modelIndx)) {
            SVL_LOG(SVL_LOG_VERBOSE, "...svlBoostedClassifierSet loading from " << modelFilename);
            _models[i] = new CvBoost();
            SVL_ASSERT(_models[i] != NULL);
            _models[i]->load(modelFilename.c_str());
        } else {
            SVL_LOG(SVL_LOG_VERBOSE, "...svlBoostedClassifierSet skipping to " << modelFilename);
        }
    }

    return true;
}

// evaluation
void svlBoostedClassifierSet::evaluate(const vector<double>& x, vector<double>& y)
{
    y.resize(_models.size());

#if 0
    CvMat *fv = cvCreateMat(1, (int)x.size(), CV_32FC1);
    for (int k = 0; k < fv->width; k++) {
        CV_MAT_ELEM(*fv, float, 0, k) = (float)x[k];
    }

    // single-threaded evaluation (only)
    for (int i = 0; i < (int)_models.size(); i++) {
        int length = cvSliceLength(CV_WHOLE_SEQ, _models[i]->get_weak_predictors());
        CvMat *weakResponses = cvCreateMat(length, 1, CV_32FC1);

        _models[i]->predict(fv, NULL, weakResponses, CV_WHOLE_SEQ);
        y[i] = cvSum(weakResponses).val[0];

        cvReleaseMat(&weakResponses);
    }

    cvReleaseMat(&fv);
#else
    // Steve's OpenCV hack to speed things up
    for (int n = 0; n < (int)_models.size(); n++) {
        CvSeq *weak = _models[n]->get_weak_predictors();

        CvSeqReader reader;
        cvStartReadSeq(weak, &reader);
        cvSetSeqReaderPos(&reader, 0);
        
        y[n] = 0.0;
        for (int i = 0; i < weak->total; i++) {
            CvBoostTree* wtree;
            CV_READ_SEQ_ELEM(wtree, reader);  

            const CvDTreeNode* node = wtree->get_root();
            while (node->left) {
                CvDTreeSplit* split = node->split;
                int vi = split->var_idx;
                int dir = x[vi] <= split->ord.c ? -1 : 1;
                if (split->inversed)
                    dir = -dir;           
                node = (dir < 0) ? node->left : node->right;
            }
            
            y[n] += node->value;
        }
    }
#endif
}

void svlBoostedClassifierSet::evaluate(const vector<vector<double> >& x,
    vector<vector<double> >& y)
{
    if (y.empty()) y.resize(x.size());
    SVL_ASSERT((x.size() == y.size()) && (!x.empty()));
    
#ifdef MULTITHREAD_EVAL
    // multi-threaded version
    for (int j = 0; j < (int)x.size(); j++) {
        y[j].resize(_models.size());
    }

    svlThreadPool threadPool(_models.size());
    threadPool.start();
    for (int i = 0; i < (int)_models.size(); i++) {
        svlBoostedClassifierEvalArgs *args = 
            new svlBoostedClassifierEvalArgs(&x, &_models, &y, i);
        threadPool.addJob(fcnEvalBoostedClassifier, args);
    }
    threadPool.finish();
#else
    // single-threaded version
    vector<CvMat *> weakResponses(_models.size(), NULL);
    for (int i = 0; i < (int)_models.size(); i++) {
        int length = cvSliceLength(CV_WHOLE_SEQ, _models[i]->get_weak_predictors());
        weakResponses[i] = cvCreateMat(length, 1, CV_32FC1);
    }

    CvMat *fv = cvCreateMat(1, (int)x[0].size(), CV_32FC1);
    for (int j = 0; j < (int)x.size(); j++) {
        for (int k = 0; k < fv->width; k++) {
            CV_MAT_ELEM(*fv, float, 0, k) = (float)x[j][k];
        }

        y[j].resize(_models.size());
        for (int i = 0; i < (int)_models.size(); i++) {
            _models[i]->predict(fv, NULL, weakResponses[i], CV_WHOLE_SEQ);
            y[j][i] = cvSum(weakResponses[i]).val[0];
        }
    }
    cvReleaseMat(&fv);

    for (int i = 0; i < (int)_models.size(); i++) {
        cvReleaseMat(&weakResponses[i]);
    }
#endif
}

// learning
double svlBoostedClassifierSet::train(const vector<vector<double> >& x, const vector<int>& y,
    int modelIndx)
{
    SVL_ASSERT((x.size() == y.size()) && (!x.empty()));

    int numRounds = getOptionAsInt("boostingRounds");
    double trimRate = getOptionAsDouble("trimRate");
    int numSplits = getOptionAsInt("numSplits");
    int pseudoCounts = getOptionAsInt("pseudoCounts");
    float classPriors[2] = {1.0f, 1.0f};

    // create data structures for training
    CvMat *data = cvCreateMat((int)x.size(), (int)x[0].size(), CV_32FC1);
    CvMat *labels = cvCreateMat((int)x.size(), 1, CV_32SC1);
    SVL_ASSERT_MSG((data != NULL) && (labels != NULL), "out of memory");

    CvMat *varType = cvCreateMat(data->width + 1, 1, CV_8UC1);
    for (int j = 0; j < data->width; j++) {
        CV_MAT_ELEM(*varType, unsigned char, j, 0) = CV_VAR_NUMERICAL;
    }
    CV_MAT_ELEM(*varType, unsigned char, data->width, 0) = CV_VAR_CATEGORICAL;

    // check input
    double maxValue = -numeric_limits<double>::max();
    double minValue = numeric_limits<double>::max();
    for (int j = 0; j < (int)x.size(); j++) {
        SVL_ASSERT(x[j].size() == x[0].size());
        for (int k = 0; k < data->width; k++) {
            cvmSet(data, j, k, x[j][k]);
            SVL_ASSERT(!isnan(x[j][k]));
            if (x[j][k] > maxValue) maxValue = x[j][k];
            if (x[j][k] < minValue) minValue = x[j][k];
        }
    }
    SVL_LOG(SVL_LOG_MESSAGE, x[0].size() << " features range from " << minValue << " to " << maxValue);
    
    // train models
    double avgRecall = 0.0;
    int numTrained = 0;
    for (int i = 0; i < (int)_models.size(); i++) {
        SVL_LOG(SVL_LOG_MESSAGE, "Training boosted classifier for label " << i << "...");
        if ((modelIndx >= 0) && (modelIndx != i)) {
            SVL_LOG(SVL_LOG_MESSAGE, "...skipping");
            continue;
        }

        // set up data vectors
        int posCount = 0;
        for (int j = 0; j < (int)y.size(); j++) {
            CV_MAT_ELEM(*labels, int, j, 0) = (y[j] == i) ? 1 : -1;
            posCount += (y[j] == i);
        }
        int negCount = y.size() - posCount;
        SVL_LOG(SVL_LOG_MESSAGE, "..." << posCount << " positive samples, " 
            << negCount << " negative samples");
            
        if (posCount == 0) {
            SVL_LOG(SVL_LOG_WARNING, "no training samples for label " << i);
            continue;
        }
            
        // train the classifier
        if (pseudoCounts > 0) {
            classPriors[0] = (float)(posCount + pseudoCounts) /
                (float)(negCount + posCount + 2 * pseudoCounts);
            classPriors[1] = (float)(negCount + pseudoCounts) / 
                (float)(negCount + posCount + 2 * pseudoCounts);
        }
        CvBoostParams parameters(CvBoost::GENTLE, numRounds, trimRate, numSplits,
            false, &classPriors[0]);
        parameters.split_criteria = CvBoost::DEFAULT;

        svlCodeProfiler::tic(svlCodeProfiler::getHandle("svlBoostedClassifierSet.boosting"));
        _models[i] = new CvBoost();
        _models[i]->train(data, CV_ROW_SAMPLE, labels, NULL, NULL, varType, NULL, parameters);
        svlCodeProfiler::toc(svlCodeProfiler::getHandle("svlBoostedClassifierSet.boosting"));
     
        // evaluate classifier (on training data)
        svlBoostedClassifierSet::svlBoostedClassifierSetStats stats = scoreModel(i, x, y);

        SVL_LOG(SVL_LOG_MESSAGE, "...done (TP, FN, TN, FP): " << stats.tp << ", " << stats.fn 
            << ", " << stats.tn << ", " << stats.fp);
        avgRecall += (double)stats.tp / (double)(stats.tp + stats.fn);
        numTrained += 1;
    }

    // free memory
    cvReleaseMat(&varType);
    cvReleaseMat(&labels);
    cvReleaseMat(&data);

    return (avgRecall / (double)numTrained);
}

double svlBoostedClassifierSet::validate(const vector<vector<double> >& x, 
    const vector<int>& y, int modelIndx)
{
    SVL_ASSERT((x.size() == y.size()) && (!x.empty()));

    // create data structures for validation
    CvMat *data = cvCreateMat((int)x.size(), (int)x[0].size(), CV_32FC1);
    CvMat *labels = cvCreateMat((int)x.size(), 1, CV_32SC1);
    SVL_ASSERT_MSG((data != NULL) && (labels != NULL), "out of memory");

    // check input
    double maxValue = -numeric_limits<double>::max();
    double minValue = numeric_limits<double>::max();
    for (int j = 0; j < (int)x.size(); j++) {
        SVL_ASSERT(x[j].size() == x[0].size());
        for (int k = 0; k < data->width; k++) {
            cvmSet(data, j, k, x[j][k]);
            SVL_ASSERT(!isnan(x[j][k]));
            if (x[j][k] > maxValue) maxValue = x[j][k];
            if (x[j][k] < minValue) minValue = x[j][k];
        }
    }
    SVL_LOG(SVL_LOG_MESSAGE, x[0].size() << " features range from " << minValue << " to " << maxValue);
    
    // validate models
    double avgRecall = 0.0;
    int numValidated = 0;
    for (int i = 0; i < (int)_models.size(); i++) {
        SVL_LOG(SVL_LOG_MESSAGE, "Validating boosted classifier for label " << i << "...");
        if ((modelIndx >= 0) && (modelIndx != i)) {
            SVL_LOG(SVL_LOG_MESSAGE, "...skipping");
            continue;
        }

        // set up data vectors
        int posCount = 0;
        for (int j = 0; j < (int)y.size(); j++) {
            CV_MAT_ELEM(*labels, int, j, 0) = (y[j] == i) ? 1 : -1;
            posCount += (y[j] == i);
        }
        int negCount = y.size() - posCount;
        SVL_LOG(SVL_LOG_MESSAGE, "..." << posCount << " positive samples, " 
            << negCount << " negative samples");
            
        svlBoostedClassifierSet::svlBoostedClassifierSetStats stats = scoreModel(i, x, y);

        SVL_LOG(SVL_LOG_MESSAGE, "...done (TP, FN, TN, FP): " << stats.tp << ", " << stats.fn 
            << ", " << stats.tn << ", " << stats.fp);
        avgRecall += (double)stats.tp / (double)(stats.tp + stats.fn);
        numValidated += 1;
    }

    // free memory
    cvReleaseMat(&labels);
    cvReleaseMat(&data);

    return (avgRecall / (double)numValidated);
}

// compute accuracy statistics for single model
svlBoostedClassifierSet::svlBoostedClassifierSetStats svlBoostedClassifierSet::scoreModel(int n,
    const vector<vector<double> >& x, const vector<int>& y)
{
    svlBoostedClassifierSet::svlBoostedClassifierSetStats stats;
    stats.tp = 0; stats.fn = 0; stats.fp = 0; stats.tn = 0;

    CvMat *fv = cvCreateMat(1, (int)x[0].size(), CV_32FC1);
    for (int j = 0; j < (int)x.size(); j++) {
        for (int k = 0; k < fv->width; k++) {
            cvmSet(fv, 0, k, x[j][k]);
        }

        float score = _models[n]->predict(fv);
        if (y[j] == n) {
            if (score > 0.0) {
                stats.tp += 1;
            } else {
                stats.fn += 1;
            }
        } else if (score > 0.0) {
            stats.fp += 1;
        } else {
            stats.tn += 1;
        }
    }

    cvReleaseMat(&fv);
    return stats;
}


