/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlGaussian.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
** DESCRIPTION:
**  See svlGaussian.h
**
** TO DO:
**  1. cache temporary variable z
**
*****************************************************************************/

#include "../base/svlCompatibility.h"

#include <cassert>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <limits>
#include <cmath>

// OpenCV libraries
#include "cv.h"
#include "cxcore.h"

// SVL libraries
#include "svlBase.h"
#include "svlGaussian.h"
#include "../vision/svlVisionUtils.h"

using namespace std;

// svlGaussian --------------------------------------------------------------

svlGaussian::svlGaussian(int n) :
    _n(n), _mu(NULL), _Sigma(NULL), _invSigma(NULL), _logZ(0.0)
{
    initialize(n);
}

svlGaussian::svlGaussian(const CvMat *mu, float sigma2) : 
    _mu(NULL), _Sigma(NULL), _invSigma(NULL)
{
    initialize(mu, sigma2);
}

svlGaussian::svlGaussian(const CvMat *mu, const CvMat *sigma2) :
    _mu(NULL), _Sigma(NULL), _invSigma(NULL)
{
    initialize(mu, sigma2);
}

svlGaussian::svlGaussian(const vector<double>& mu, float sigma2) :
    _mu(NULL), _Sigma(NULL), _invSigma(NULL)
{
    initialize(mu, sigma2);
}

svlGaussian::svlGaussian(const svlSufficientStats& stats) :
    _n(stats.size()), _mu(NULL), _Sigma(NULL), _invSigma(NULL), _logZ(0.0)
{
    initialize(_n);
    train(stats, SVL_EPSILON);
}

svlGaussian::svlGaussian(const svlGaussian& model) : 
    _n(model._n), _logZ(model._logZ)
{
    _mu = cvCloneMat(model._mu);
    _Sigma = cvCloneMat(model._Sigma);
    _invSigma = cvCloneMat(model._invSigma);
}

svlGaussian::~svlGaussian()
{
    freeStorage();
}

// initialization
void svlGaussian::initialize(int n)
{
    SVL_ASSERT(n > 0);

    _n = n;
    freeStorage();
    _mu = cvCreateMat(_n, 1, CV_32FC1);
    _Sigma = cvCreateMat(_n, _n, CV_32FC1);
    
    cvZero(_mu);
    cvSetIdentity(_Sigma, cvScalar(1.0f));
    _invSigma = cvCloneMat(_Sigma);

    _logZ = -0.5f * log(2.0 * M_PI);    
}

void svlGaussian::initialize(const CvMat *mu, float sigma2)
{
    SVL_ASSERT((mu != NULL) && (sigma2 > 0.0f));

    _n = mu->rows;
    SVL_ASSERT(mu->cols == 1);

    freeStorage();
    _mu = cvCloneMat(mu);
    _Sigma = cvCreateMat(_n, _n, CV_32FC1);
    cvSetIdentity(_Sigma, cvScalar(sigma2));
    _invSigma = cvCreateMat(_n, _n, CV_32FC1);
    cvSetIdentity(_Sigma, cvScalar(1.0f / sigma2));

    _logZ = -0.5f * (float)_n * log(2.0 * M_PI) - 0.5f * log(cvDet(_Sigma));
}

void svlGaussian::initialize(const CvMat *mu, const CvMat *sigma2)
{
    SVL_ASSERT((mu != NULL) && (sigma2 != NULL));

    _n = mu->rows;
    SVL_ASSERT(mu->cols == 1);
    SVL_ASSERT((sigma2->rows == _n) && (sigma2->cols == _n));

    freeStorage();
    _mu = cvCloneMat(mu);
    _Sigma = cvCloneMat(sigma2);
    updateCachedParameters();
}

void svlGaussian::initialize(const vector<double>& mu, float sigma2)
{
    SVL_ASSERT(!mu.empty());

    _n = (int)mu.size();
    
    freeStorage();
    _mu = cvCreateMat(_n, 1, CV_32FC1);
    for (int i = 0; i < _n; i++) {
        CV_MAT_ELEM(*_mu, float, _n, 0) = (float)mu[i];
    }

    _Sigma = cvCreateMat(_n, _n, CV_32FC1);
    cvSetIdentity(_Sigma, cvScalar(sigma2));
    _invSigma = cvCreateMat(_n, _n, CV_32FC1);
    cvSetIdentity(_Sigma, cvScalar(1.0f / sigma2));

    _logZ = -0.5f * (float)_n * log(2.0 * M_PI) - 0.5f * log(cvDet(_Sigma));    
}

// conditioning
svlGaussian svlGaussian::reduce(const vector<double>& x2,
    const vector<int>& indx2) const
{
    // TODO: cache g._Sigma and sigmaProduct so that repeated calls
    // with the same indx2 doesn't require additional computation

    SVL_ASSERT((x2.size() == indx2.size()) && (x2.size() < (unsigned)_n));
    
    int n1 = _n - (int)x2.size();
    int n2 = (int)x2.size();
    vector<int> indxAll(_n);
    for (int i = 0; i < _n; i++) {
        indxAll[i] = i;
    }
    vector<int> indx1;
    set_difference(indxAll.begin(), indxAll.end(),
        indx2.begin(), indx2.end(), 
        insert_iterator<vector<int> >(indx1, indx1.end()));

    svlGaussian g(n1);        
    
    CvMat *mu2 = cvCreateMat(n2, 1, CV_32FC1);
    CvMat *Sigma22 = cvCreateMat(n2, n2, CV_32FC1);
    CvMat *Sigma21 = cvCreateMat(n2, n1, CV_32FC1);

    for (int i = 0; i < n2; i++) {
        SVL_ASSERT((indx2[i] >= 0) && (indx2[i] < _n));
        CV_MAT_ELEM(*mu2, float, i, 0) = (float)x2[i] - 
            CV_MAT_ELEM(*_mu, float, indx2[i], 0);
        for (int j = 0; j < n2; j++) {
            CV_MAT_ELEM(*Sigma22, float, i, j) = 
                CV_MAT_ELEM(*_Sigma, float, indx2[i], indx2[j]);
        }
        for (int j = 0; j < n1; j++) {
            CV_MAT_ELEM(*Sigma21, float, i, j) = 
                CV_MAT_ELEM(*_Sigma, float, indx2[i], indx1[j]);
        }
    }

    CvMat *sigmaProduct = cvCreateMat(n2, n1, CV_32FC1);
    cvSolve(Sigma22, Sigma21, sigmaProduct);
    
    cvGEMM(Sigma21, sigmaProduct, -1.0, NULL, 1.0, g._Sigma, CV_GEMM_A_T);
    cvGEMM(sigmaProduct, mu2, 1.0, NULL, 1.0, g._mu, CV_GEMM_A_T);

    for (int i = 0; i < n1; i++) {
        CV_MAT_ELEM(*g._mu, float, i, 0) += CV_MAT_ELEM(*_mu, float, indx1[i], 0);
        for (int j = 0; j < n1; j++) {
            CV_MAT_ELEM(*g._Sigma, float, i, j) +=
                CV_MAT_ELEM(*_Sigma, float, indx1[i], indx1[j]);
        }
    }

    cvReleaseMat(&sigmaProduct);
    cvReleaseMat(&Sigma21);
    cvReleaseMat(&Sigma22);
    cvReleaseMat(&mu2);

    g.updateCachedParameters();
    return g;
}

svlGaussian svlGaussian::reduce(const map<int, double>& x) const
{
    SVL_ASSERT(false);
    return svlGaussian();
}

svlConditionalGaussian svlGaussian::conditionOn(const vector<int>& indx2) const
{
    SVL_ASSERT(indx2.size() < (unsigned)_n);
    
    int n1 = _n - (int)indx2.size();
    int n2 = (int)indx2.size();
    vector<int> indxAll(_n);
    for (int i = 0; i < _n; i++) {
        indxAll[i] = i;
    }
    vector<int> indx1;
    set_difference(indxAll.begin(), indxAll.end(),
        indx2.begin(), indx2.end(), 
        insert_iterator<vector<int> >(indx1, indx1.end()));

    CvMat *mu2 = cvCreateMat(n2, 1, CV_32FC1);
    CvMat *Sigma22 = cvCreateMat(n2, n2, CV_32FC1);
    CvMat *Sigma21 = cvCreateMat(n2, n1, CV_32FC1);

    for (int i = 0; i < n2; i++) {
        SVL_ASSERT((indx2[i] >= 0) && (indx2[i] < _n));
        CV_MAT_ELEM(*mu2, float, i, 0) = CV_MAT_ELEM(*_mu, float, indx2[i], 0);
        for (int j = 0; j < n2; j++) {
            CV_MAT_ELEM(*Sigma22, float, i, j) = 
                CV_MAT_ELEM(*_Sigma, float, indx2[i], indx2[j]);
        }
        for (int j = 0; j < n1; j++) {
            CV_MAT_ELEM(*Sigma21, float, i, j) = 
                CV_MAT_ELEM(*_Sigma, float, indx2[i], indx1[j]);
        }
    }

    CvMat *mu = cvCreateMat(n1, 1, CV_32FC1);
    CvMat *Sigma = cvCreateMat(n1, n1, CV_32FC1);
    CvMat *sigmaProduct = cvCreateMat(n2, n1, CV_32FC1);
    cvSolve(Sigma22, Sigma21, sigmaProduct);
    
    cvGEMM(Sigma21, sigmaProduct, -1.0, NULL, 1.0, Sigma, CV_GEMM_A_T);
    cvGEMM(sigmaProduct, mu2, -1.0, NULL, 1.0, mu, CV_GEMM_A_T);

    for (int i = 0; i < n1; i++) {
        CV_MAT_ELEM(*mu, float, i, 0) += CV_MAT_ELEM(*_mu, float, indx1[i], 0);
        for (int j = 0; j < n1; j++) {
            CV_MAT_ELEM(*Sigma, float, i, j) +=
                CV_MAT_ELEM(*_Sigma, float, indx1[i], indx1[j]);
        }
    }

    CvMat *sigmaGain = cvCreateMat(n1, n2, CV_32FC1);
    cvTranspose(sigmaProduct, sigmaGain);

    svlConditionalGaussian g(mu, Sigma, sigmaGain);

    cvReleaseMat(&sigmaGain);
    cvReleaseMat(&mu);
    cvReleaseMat(&Sigma);
    cvReleaseMat(&sigmaProduct);
    cvReleaseMat(&Sigma21);
    cvReleaseMat(&Sigma22);
    cvReleaseMat(&mu2);

    return g;
}

// i/o
bool svlGaussian::save(const char *filename) const
{
    ofstream ofs(filename);
    if (ofs.fail()) {
	return false;
    }

    ofs << _n << "\n\n";
    for (int i = 0; i < _n; i++) {
        if (i > 0) ofs << " ";
        ofs << setprecision(16) << CV_MAT_ELEM(*_mu, float, i, 0);
    }
    ofs << "\n\n";

    for (int i = 0; i < _n; i++) {
        for (int j = 0; j < _n; j++) {
            if (j > 0) ofs << " ";
            ofs << setprecision(16) << CV_MAT_ELEM(*_Sigma, float, i, j);
        }
        ofs << "\n";
    }

    ofs.close();
    return true;    
}

bool svlGaussian::load(const char *filename)
{
    ifstream ifs(filename);
    if (ifs.fail()) {
	return false;
    }

    ifs >> _n;
    initialize(_n);

    for (int i = 0; i < _n; i++) {
        ifs >> CV_MAT_ELEM(*_mu, float, i, 0);
    }
    SVL_ASSERT(!ifs.fail());

    for (int i = 0; i < _n; i++) {
        for (int j = 0; j < _n; j++) {
            ifs >> CV_MAT_ELEM(*_Sigma, float, i, j);
        }
    }
    SVL_ASSERT(!ifs.fail());
    ifs.close();

    updateCachedParameters();
    return true;
}

// evaluate (log-likelihood)
void svlGaussian::evaluate(const CvMat *x, CvMat *p) const
{
    SVL_ASSERT((x != NULL) && (p != NULL));
    SVL_ASSERT((x->cols == _n) && (x->rows == p->rows) && (p->cols == 1));

    CvMat *z = cvCreateMat(_n, 1, CV_32FC1);
    for (int i = 0; i < x->rows; i++) {
        for (int j = 0; j < _n; j++) {
            CV_MAT_ELEM(*z, float, j, 0) = CV_MAT_ELEM(*x, float, i, j);
        }
        double s = cvMahalonobis(z, _mu, _invSigma);
        CV_MAT_ELEM(*p, float, i, 0) = -0.5f * s * s + _logZ;
    }
    cvReleaseMat(&z);
}

void svlGaussian::evaluate(const vector<vector<double> >& x, vector<double>& p) const
{
    SVL_ASSERT(x.size() == p.size());

    CvMat *z = cvCreateMat(_n, 1, CV_32FC1);
    for (int i = 0; i < (int)x.size(); i++) {
        SVL_ASSERT(x[i].size() == (unsigned)_n);
        for (int j = 0; j < _n; j++) {
            CV_MAT_ELEM(*z, float, j, 0) = (float)x[i][j];
        }
        double s = cvMahalonobis(z, _mu, _invSigma);
        p[i] = -0.5f * s * s + _logZ;
    }
    cvReleaseMat(&z);
}

double svlGaussian::evaluateSingle(const CvMat *x) const
{
    SVL_ASSERT_MSG((x != NULL) && ((x->rows == _n) || (x->cols == _n)) &&
        ((x->rows == 1) || (x->cols == 1)), x->rows << " x " << x->cols << " != " << _n);

    CvMat *z = cvCreateMat(_n, 1, CV_32FC1);
    if (x->rows == _n) {
        cvCopy(x, z);
    } else {
        cvTranspose(x, z);
    }

    double s = cvMahalonobis(z, _mu, _invSigma);
    cvReleaseMat(&z);

    return -0.5f * s * s + _logZ;
}

double svlGaussian::evaluateSingle(const vector<double>& x) const
{
    SVL_ASSERT(x.size() == (unsigned)_n);

    CvMat *z = cvCreateMat(_n, 1, CV_32FC1);
    for (int i = 0; i < _n; i++) {
        CV_MAT_ELEM(*z, float, i, 0) = (float)x[i];
    }

    double s = cvMahalonobis(z, _mu, _invSigma);
    cvReleaseMat(&z);

    return -0.5f * s * s + _logZ;
}

double svlGaussian::evaluateSingle(double x) const
{
    SVL_ASSERT(_n == 1);
    
    double z = (x - CV_MAT_ELEM(*_mu, float, 0, 0));
    double s = CV_MAT_ELEM(*_invSigma, float, 0, 0) * z * z;
    return -0.5f * s + _logZ;
}

// learn parameters
void svlGaussian::train(const CvMat *x, float lambda)
{
    SVL_ASSERT((x != NULL) && (x->cols == _n) && 
        ((x->rows > 1) || (lambda > 0.0)));

    cvZero(_mu);
    cvZero(_Sigma);

    // accumulate statistics
    for (int i = 0; i < x->rows; i++) {
        for (int j = 0; j < _n; j++) {
            CV_MAT_ELEM(*_mu, float, j, 0) += CV_MAT_ELEM(*x, float, i, j);
            for (int k = 0; k <= j; k++) {
                CV_MAT_ELEM(*_Sigma, float, j, k) +=
                    CV_MAT_ELEM(*x, float, i, j) * CV_MAT_ELEM(*x, float, i, k);
            }
        }
    }

    // normalize
    for (int j = 0; j < _n; j++) {
        CV_MAT_ELEM(*_mu, float, j, 0) /= (float)x->rows;
        for (int k = 0; k <= j; k++) {
            CV_MAT_ELEM(*_Sigma, float, j, k) /= (float)x->rows;
            CV_MAT_ELEM(*_Sigma, float, j, k) -= 
                CV_MAT_ELEM(*_mu, float, j, 0) * CV_MAT_ELEM(*_mu, float, k, 0);
        }
        CV_MAT_ELEM(*_Sigma, float, j, j) += lambda;
    }

    // copy symmetric part
    for (int j = 0; j < _n; j++) {
        for (int k = j + 1; k < _n; k++) {
            CV_MAT_ELEM(*_Sigma, float, j, k) =
                CV_MAT_ELEM(*_Sigma, float, k, j);
        }
    }

    updateCachedParameters();
}

void svlGaussian::train(const vector<vector<double> >& x, float lambda)
{
    SVL_ASSERT((x.size() > 1) || (lambda > 0.0));

    cvZero(_mu);
    cvZero(_Sigma);

    // accumulate statistics
    for (int i = 0; i < (int)x.size(); i++) {
        SVL_ASSERT(x[i].size() == (unsigned)_n);
        for (int j = 0; j < _n; j++) {
            CV_MAT_ELEM(*_mu, float, j, 0) += (float)x[i][j];
            for (int k = 0; k <= j; k++) {
                CV_MAT_ELEM(*_Sigma, float, j, k) += (float)(x[i][j] * x[i][k]);
            }
        }
    }

    // normalize
    for (int j = 0; j < _n; j++) {
        CV_MAT_ELEM(*_mu, float, j, 0) /= (float)x.size();
        for (int k = 0; k <= j; k++) {
            CV_MAT_ELEM(*_Sigma, float, j, k) /= (float)x.size();
            CV_MAT_ELEM(*_Sigma, float, j, k) -= 
                CV_MAT_ELEM(*_mu, float, j, 0) * CV_MAT_ELEM(*_mu, float, k, 0);
        }
        CV_MAT_ELEM(*_Sigma, float, j, j) += lambda;
    }

    // copy symmetric part
    for (int j = 0; j < _n; j++) {
        for (int k = j + 1; k < _n; k++) {
            CV_MAT_ELEM(*_Sigma, float, j, k) =
                CV_MAT_ELEM(*_Sigma, float, k, j);
        }
    }

    updateCachedParameters();
}

void svlGaussian::train(const svlSufficientStats& stats, float lambda)
{
    SVL_ASSERT_MSG(stats.size() == _n, stats.size() << " != " << _n);

    // normalize sufficient statistics
    double nu = 1.0 / stats.count();
    for (int j = 0; j < _n; j++) {
        CV_MAT_ELEM(*_mu, float, j, 0) = (float)(nu * stats.sum(j));
        for (int k = 0; k <= j; k++) {
            CV_MAT_ELEM(*_Sigma, float, j, k) = (float)(nu * stats.sum2(j, k));
            CV_MAT_ELEM(*_Sigma, float, j, k) -= 
                CV_MAT_ELEM(*_mu, float, j, 0) * CV_MAT_ELEM(*_mu, float, k, 0);
        }

        CV_MAT_ELEM(*_Sigma, float, j, j) += lambda;
    }

    // copy symmetric part
    for (int j = 0; j < _n; j++) {
        for (int k = j + 1; k < _n; k++) {
            CV_MAT_ELEM(*_Sigma, float, j, k) =
                CV_MAT_ELEM(*_Sigma, float, k, j);
        }
    }

    updateCachedParameters();
}

// kl-divergence, KL(*this || model)
// TODO: test
double svlGaussian::klDivergence(const svlGaussian& model) const
{
    SVL_ASSERT(model._n == _n);
    
    double s = cvMahalonobis(model._mu, _mu, model._invSigma);
    double d = _logZ - model._logZ + 
        0.5 * (cvDotProduct(model._invSigma, _Sigma) + s * s - (double)_n);
    
    return d;
}

double svlGaussian::klDivergence(const svlSufficientStats& stats) const
{
    SVL_ASSERT(stats.size() == _n);
    return klDivergence(svlGaussian(stats));
}

// standard operators
svlGaussian& svlGaussian::operator=(const svlGaussian& model)
{
    freeStorage();
    _n = model._n;
    _mu = cvCloneMat(model._mu);
    _Sigma = cvCloneMat(model._Sigma);
    _invSigma = cvCloneMat(model._invSigma);
    _logZ = model._logZ;

    return *this;
}

// protected functions
void svlGaussian::freeStorage()
{
    if (_invSigma != NULL) cvReleaseMat(&_invSigma);
    if (_Sigma != NULL) cvReleaseMat(&_Sigma);
    if (_mu != NULL) cvReleaseMat(&_mu);
}

void svlGaussian::updateCachedParameters()
{
    if (_invSigma != NULL) cvReleaseMat(&_invSigma);

    _invSigma = cvCreateMat(_n, _n, CV_32FC1);
    double det = cvInvert(_Sigma, _invSigma, CV_LU);
    if (svlLogger::getLogLevel() >= SVL_LOG_DEBUG) {
        SVL_LOG(SVL_LOG_DEBUG, "svlGaussian::_Sigma =");
        dump(*_Sigma);
    }
    SVL_ASSERT_MSG(isfinite(det) && (det > 0.0), "|Sigma| = " << det);
    _logZ = -0.5f * (float)_n * log(2.0 * M_PI) - 0.5f * log(det);
}

// svlConditionalGaussian ---------------------------------------------------------

svlConditionalGaussian::svlConditionalGaussian(const CvMat *mu, const CvMat *Sigma, 
    const CvMat *SigmaGain) :
    _mu(NULL), _Sigma(NULL), _SigmaGain(NULL), _tmp(NULL)
{
    SVL_ASSERT((mu != NULL) && (Sigma != NULL) && (SigmaGain != NULL));
    _n = SigmaGain->rows;
    _m = SigmaGain->cols;
    SVL_ASSERT(mu->rows == _n);
    SVL_ASSERT((Sigma->rows == _n) && (Sigma->cols == _n));
    
    _mu = cvCloneMat(mu);
    _Sigma = cvCloneMat(Sigma);
    _SigmaGain = cvCloneMat(SigmaGain);
    _tmp = cvCreateMat(_n, 1, CV_32FC1);
}

svlConditionalGaussian::svlConditionalGaussian(const svlConditionalGaussian& model) :
    _n(model._n), _m(model._m), _mu(NULL), _Sigma(NULL), _SigmaGain(NULL), _tmp(NULL)
{
    _mu = cvCloneMat(model._mu);
    _Sigma = cvCloneMat(model._Sigma);
    _SigmaGain = cvCloneMat(model._SigmaGain);    
    _tmp = cvCreateMat(_n, 1, CV_32FC1);
}

svlConditionalGaussian::~svlConditionalGaussian()
{
    if (_mu != NULL) cvReleaseMat(&_mu);
    if (_Sigma != NULL) cvReleaseMat(&_Sigma);
    if (_SigmaGain != NULL) cvReleaseMat(&_SigmaGain);
    if (_tmp != NULL) cvReleaseMat(&_tmp);
}

// i/o
bool svlConditionalGaussian::save(const char *filename) const
{
    ofstream ofs(filename);
    if (ofs.fail()) {
	return false;
    }

    ofs << _n << " " << _m << "\n\n";

    // mu
    for (int i = 0; i < _n; i++) {
        if (i > 0) ofs << " ";
        ofs << CV_MAT_ELEM(*_mu, float, i, 0);
    }
    ofs << "\n\n";

    // Sigma
    for (int i = 0; i < _n; i++) {
        for (int j = 0; j < _n; j++) {
            if (j > 0) ofs << " ";
            ofs << CV_MAT_ELEM(*_Sigma, float, i, j);
        }
        ofs << "\n";
    }
    ofs << "\n";

    // SigmaGain
    for (int i = 0; i < _n; i++) {
        for (int j = 0; j < _m; j++) {
            if (j > 0) ofs << " ";
            ofs << CV_MAT_ELEM(*_SigmaGain, float, i, j);
        }
        ofs << "\n";
    }

    ofs.close();
    return true;    
}

bool svlConditionalGaussian::load(const char *filename)
{
    ifstream ifs(filename);
    if (ifs.fail()) {
	return false;
    }

    if (_mu != NULL) cvReleaseMat(&_mu);
    if (_Sigma != NULL) cvReleaseMat(&_Sigma);
    if (_SigmaGain != NULL) cvReleaseMat(&_SigmaGain);
    if (_tmp != NULL) cvReleaseMat(&_tmp);

    ifs >> _n >> _m;

    _mu = cvCreateMat(_n, 1, CV_32FC1);
    _Sigma = cvCreateMat(_n, _n, CV_32FC1);
    _SigmaGain = cvCreateMat(_n, _m, CV_32FC1);
    _tmp = cvCreateMat(_n, 1, CV_32FC1);

    for (int i = 0; i < _n; i++) {
        ifs >> CV_MAT_ELEM(*_mu, float, i, 0);
    }
    SVL_ASSERT(!ifs.fail());

    for (int i = 0; i < _n; i++) {
        for (int j = 0; j < _n; j++) {
            ifs >> CV_MAT_ELEM(*_Sigma, float, i, j);
        }
    }
    SVL_ASSERT(!ifs.fail());

    for (int i = 0; i < _n; i++) {
        for (int j = 0; j < _m; j++) {
            ifs >> CV_MAT_ELEM(*_SigmaGain, float, i, j);
        }
    }
    SVL_ASSERT(!ifs.fail());
    ifs.close();

    return true;
}

svlGaussian svlConditionalGaussian::reduce(const CvMat *x)
{
    SVL_ASSERT((x != NULL) && (x->rows == _m) && (x->cols == 1));
    cvMatMulAdd(_SigmaGain, x, _mu, _tmp);
    return svlGaussian(_tmp, _Sigma);
}

svlGaussian svlConditionalGaussian::reduce(const vector<double>& x)
{
    CvMat *xm = cvCreateMat(x.size(), 1, CV_32FC1);
    for (int i = 0; i < (int)x.size(); i++) {
        CV_MAT_ELEM(*xm, float, i, 0) = (float)x[i];
    }

    svlGaussian g = reduce(xm);
    cvReleaseMat(&xm);
    
    return g;
}
