/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlPatchDictionary.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Sid Batra <sidbatra@cs.stanford.edu>
**              David Breeden <breeden@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
**
*****************************************************************************/


#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <list>
#include <limits>
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlVision.h"
#include "svlDevel.h"


// speed up by using threads on a linux machine

#define USE_THREADS

#define TOTAL_THREADS 8

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#undef max
#undef USE_THREADS
#endif

#ifdef USE_THREADS
#include <pthread.h>
#endif

using namespace std;


// Threading functions -------------------------------------------------------

struct NormalizePatchArgs {
  NormalizePatchArgs(const svlPatchDictionary * par, IplImage* const p, const int i) : parent(par), patch(p), index(i) {}

  const svlPatchDictionary * parent;
  IplImage* const patch;
  const int index;
};

void * normalizePatch(void *argsP, unsigned tid) {
  NormalizePatchArgs *args = (NormalizePatchArgs*) argsP;
  IplImage* const patch = args->patch;

  vector<float> depths(patch->width * patch->height);
  int i = 0;
  for (int y = 0; y < patch->height; y++) {
      for (int x = 0; x < patch->width; x++) {
          depths[i++] = CV_IMAGE_ELEM(patch, float, y, x);
      }
  }

  sort(depths.begin(), depths.end());

  const float median = depths[depths.size()/2];

  for( int y=0; y<patch->height; y++) {
    for( int x=0; x<patch->width; x++) {
      CV_IMAGE_ELEM(patch, float, y, x) -= median;
    }
  }

  args->parent->_normalizedDepthPatches[args->index] = patch;

  delete args;

  return NULL;
}


struct CalcFVArgs {
  CalcFVArgs(const IplImage *i,
	     const IplImage *sI,
	     const IplImage *sI2,
	     vector<IplImage **> *r,
	     bool s, 
	     vector<svlSmartPointer<svlPatchDefinition> > *e,
	     const int eI, const vector<CvPoint> *w,
	     vector<vector<double> > *fv, unsigned outputOffset) :
    image(i), sumImage(sI), sumImage2(sI2), absResponses(r),
    sparse(s), entries(e), entryIndex(eI), windows(w), v(fv), outputOffset(outputOffset) {}
  const IplImage *image;
  const IplImage *sumImage;
  const IplImage *sumImage2;
  vector<IplImage **> *absResponses;
  const bool sparse;
  vector<svlSmartPointer<svlPatchDefinition> > *entries;
  const int entryIndex;
  const vector<CvPoint> *windows;
  vector<vector<double> > *v;
  unsigned outputOffset;
};

void *calcFV(void *argP, unsigned tid) 
{
  CalcFVArgs *args = (CalcFVArgs*) argP;

  const int entryIndex = args->entryIndex;
  svlPatchDefinition *entry = args->entries->at(entryIndex);

  bool use_integral_images = (args->sumImage != NULL && args->sumImage2 != NULL);
  bool use_abs_responses = (args->absResponses != NULL) && (args->absResponses->size() != 0);
  bool sparse = args->sparse;

  // if not sparse, then want to first compute the response image and then call
  // patchValue to extract the highest response within the appropriate neighborhood
  if (!sparse && entry->patchType() != SVL_DEPTH_PATCH) {
  //if (!sparse) {
      IplImage *response = NULL;

      if (use_integral_images  && (entry->patchType() == SVL_INTENSITY_PATCH)) {
		  response = ((svlIntensityPatchDefinition*)entry)->responseImage(args->image, args->sumImage, args->sumImage2);
      } else if (use_abs_responses && (entry->patchType() == SVL_ABS_PATCH)) {
	svlAbsPatchDefinition *patch = (svlAbsPatchDefinition *)(entry);
	svlAbsPatchType t = patch->getAbsType();
	response = patch->responseImage(args->absResponses->at(t));
      } else {
	response = entry->responseImage(args->image);
      }

      assert(response);

      const vector<CvPoint> *windows = args->windows;
      for (unsigned j = 0; j < windows->size(); j++) {
	args->v->at(j)[entryIndex+args->outputOffset] = entry->patchValue(response, windows->at(j));
      }

      cvReleaseImage(&response);
  }

  // if sparse, want to do as few computations as possible, in particular avoiding making
  // the entire response image (this is not currently implemented as such for abs patches for example,
  // but in general the patchValues function is reserved for such sparse computations)
  else {
      vector<double> p;

      if (use_abs_responses && (entry->patchType() == SVL_ABS_PATCH)) {
	svlAbsPatchDefinition *patch = (svlAbsPatchDefinition *)(entry);
	svlAbsPatchType t = patch->getAbsType();
	p = patch->patchValues(args->absResponses->at(t), *(args->windows));
      } else {
	p  = entry->patchValues(args->image, *(args->windows));
      }

      for (unsigned j = 0; j < args->windows->size(); j++) {
	args->v->at(j)[entryIndex + args->outputOffset ] = p[j];
      }
  }

  delete args;

  return NULL;
}


// svlPatchDictionary class --------------------------------------------------


svlPatchDictionary::svlPatchDictionary(unsigned width, unsigned height) : 
  _bUseIntImg(USE_INTEGRAL_IMAGES_DEFAULT)
{
    _windowSize = cvSize(width, height);
}

svlPatchDictionary::svlPatchDictionary(const svlPatchDictionary& d) :
  _entries(d._entries), _bUseIntImg(USE_INTEGRAL_IMAGES_DEFAULT)
{
  _windowSize = d._windowSize;


  for (unsigned i = 0; i < _entries.size(); i++)
    _entries[i]->setParent(this);

  for (unsigned i = 0; i < d._normalizedDepthPatches.size(); i++)
    _normalizedDepthPatches.push_back(cvCloneImage(d._normalizedDepthPatches[i]));


  if (d._responses.size() || d._response_origin.size())
    {
      SVL_LOG(SVL_LOG_FATAL, "Abs patches are not handled correctly by the patch dictionary copy constructor");
    }

    // do nothing
}

svlPatchDictionary::~svlPatchDictionary()
{
    // do nothing
}

void svlPatchDictionary::clear()
{
    _entries.clear();
}

void svlPatchDictionary::truncate(unsigned n)
{
    _entries.resize(n);
}

void svlPatchDictionary::filterEntries(const vector<bool> &toKeep)
{
  SVL_ASSERT_MSG(toKeep.size() == _entries.size(), \
		 "The mask passed in must have the same number of entries as are present in the dictionary");

  for(int i = (int)_entries.size() - 1; i >= 0; i--) {
    if (!toKeep[i])
      _entries.erase(_entries.begin() + i);
  }
}

void svlPatchDictionary::alterValidChannel(int newChannel)
{
    for (unsigned i=0 ; i<_entries.size() ; i++)
    {
        _entries[i]->setValidChannel(newChannel);
    }
}

void svlPatchDictionary::mergeDictionary( svlPatchDictionary input )
{
    for (unsigned i=0 ; i<input.numFeatures() ; i++)
    {
        _entries.push_back( input._entries[i] );
    }
}

bool svlPatchDictionary::read(const char *filename)
{
  XMLNode root = XMLNode::parseFile(filename);//don't specify the desired
  //root name-- otherwise if there is a patch dictionary embedded in a 
  //a different XML node this will find that node and pretend it was the root
  //We don't want that to happen, because then this method could be passed
  //a composite feature extractor on accident and would not trigger an error
    bool rval = read(root);
    if(!rval)
      SVL_LOG(SVL_LOG_WARNING, "could not read patch dictionary file " << filename);
    return rval;
}

pair<int, int> svlPatchDictionary::readClassifierSize(const char *filename) {
  XMLNode root = XMLNode::parseFile(filename, "PatchDictionary");

  pair<int, int> result;
  result.first = atoi(root.getAttribute("width"));
  result.second = atoi(root.getAttribute("height"));

  return result;
}

bool svlPatchDictionary::read(XMLNode &root)
{

    _entries.clear();

    if (root.isEmpty()) {
        SVL_LOG(SVL_LOG_WARNING, "Patch dictionary file is missing or has incorrect root.");
	return false;
    }

    if (string(root.getName()) != "PatchDictionary" ) {
        SVL_LOG(SVL_LOG_WARNING, "Attempt to read patch dictionary from XMLNode that is not a patch dictionary.");
	return false;
    }

    _windowSize.width = atoi(root.getAttribute("width"));
    _windowSize.height = atoi(root.getAttribute("height"));
    int n = atoi(root.getAttribute("numEntries"));


    assert(root.nChildNode("PatchDefinition") == n);
    _entries.resize(n, NULL);

    for (unsigned i = 0; i < _entries.size(); i++) 
    {		
        XMLNode node = root.getChildNode("PatchDefinition", i);
        _entries[i] = svlPatchDefinition::createPatchDefinition(node, this);
        assert(_entries[i] != NULL);
    }



    return true;
}

bool svlPatchDictionary::writeOut(ofstream &ofs)
{
    if (ofs.fail()) { return false; }

    ofs << "<PatchDictionary version=\"1\"\n"
        << "  width=\"" <<  _windowSize.width << "\"\n"
        << "  height=\"" << _windowSize.height << "\"\n"
        << "  numEntries=\"" << _entries.size() << "\">\n";

    for (unsigned i = 0; i < _entries.size(); i++) {
        _entries[i]->write(ofs);
    }

    ofs << "</PatchDictionary>\n";

    return true;
}

//Will not place rectangle anywhere where mask contains 1s
static void sampleLocationsRandomly(int minWidth, int minHeight,
				    int maxWidth, int maxHeight,
				    int windowWidth, int windowHeight,
				    unsigned n, CvRNG &rngState, 
				    vector<CvRect> &rectangles,
				    CvMat * mask = NULL) {

  if (mask)
    {
      assert(windowWidth == mask->cols);
      assert(windowHeight == mask->rows);
    }

  rectangles.clear();
  for (unsigned j = 0; j < n; j++) {
    CvRect r;

    bool accepted = false;;
    int triesRemaining = 10;

    while ( (!accepted) && (triesRemaining > 0) )
      {
	r.width = (cvRandInt(&rngState) % (maxWidth - minWidth)) + minWidth;
	r.height = (cvRandInt(&rngState) % (maxHeight - minHeight)) + minHeight;
	r.x = cvRandInt(&rngState) % (windowWidth - r.width);
	r.y = cvRandInt(&rngState) % (windowHeight - r.height);

	if (mask)
	  {
	    triesRemaining--;

	    accepted = true;

	    for (int y = r.y; accepted && y < r.y + r.height; y++)
	      {
		for (int x = r.x; x < r.x + r.width; x++)
		  {
		    if (cvmGet(mask,y,x) > 0.5)
		      {
			accepted = false;
			break;
		      }
		  }
	      }
	    
	  }
	else
	  {
	    accepted = true;
	  }
      }
    
    if (accepted)
      rectangles.push_back(r);
  } // patches
}

/*********************
 * TODO: finish this function (quite useful potentially)
 *

static void sampleInterestPoints(const IplImage *image,
				 int minWidth, int minHeight,
				 int maxWidth, int maxHeight,
				 unsigned n, vector<cvRect> &rectangles) {

  SVL_ASSERT_MSG(image->nChannels == 1, "need to implement conversion to grayscale in buildDictionary");

  IplImage *eig_image = cvCreateImage(cvRect(image->width, image->height),
				      IPL_DEPTH_32F, 1);
  IplImage *temp_image = cvCreateImage(cvRect(image->width, image->height),
				       IPL_DEPTH_32F, 1);

  int corner_count = 200;
  CvPoint2D32f *corners = new CvPoint2D32f[corner_count];
  double min_distance = 3;

  for(double quality_level = 0.1; quality_level <= 1; quality_level += 0.1) {
    cvGoodFeaturesToTrack(image, eig_image, temp_image, corners,
			  &corner_count, quality_level, min_distance);
    
    cerr << quality_level << " " << corner_count << endl;
  }
  }*/



// for the other two patch types we were assuming that the images 
// are greyscale, and they are passed in as such; the channels are
// represented by the vector of images in samples[i]. for
// AbsPatches, we can pass in multi-channeled images,
void svlPatchDictionary::buildDictionary(
const vector<vector<IplImage *> >& samples,
const vector<svlPatchDefinitionType>& imageTypes, 
unsigned n,
bool interestPoints, const vector<CvMat *> * masks)
{
    assert(imageTypes.size() == samples[0].size());

    CvRNG rngState(0xffffffff);

    // to suppress warnings
    svlAbsPatchType type = SVL_ABS_NUM;

    int minWidth_absPatch[SVL_ABS_NUM];
    int minHeight_absPatch[SVL_ABS_NUM];
    
    for (int i = 0; i < SVL_ABS_NUM; i++) {
      svlAbsPatchDefinition s( this, (svlAbsPatchType)i);
      minWidth_absPatch[i] = s.getMinWidth();
      minHeight_absPatch[i] = s.getMinHeight();
    }

    int numChannels = (int)samples[0].size();    

    unsigned numDepthAdded = 0;
    unsigned numDepthSubtracted = 0;

    for (int c = 0; c < numChannels; c++) {
      for (unsigned i = 0; i < samples.size(); i++) {
	assert(samples[i].size() == (unsigned)numChannels);
	if (samples[i][c] == NULL)
	  continue;
	assert((samples[i][c]->width == _windowSize.width) && 
	       (samples[i][c]->height == _windowSize.height));
	
	// if this image is uniform, then no patches extracted from it will be valid
	//if (isImageUniform(samples[i][c])) continue;

	// minimum patch size is MIN(4, _windowSize / 8)
	// maximum patch size is _windowSize / 2
	int minWidth = MIN(4, _windowSize.width / 8);
	int minHeight = MIN(4, _windowSize.height / 8);

	if (imageTypes[c] == SVL_ABS_PATCH) {
	  type = (svlAbsPatchType)(cvRandInt(&rngState) % SVL_ABS_NUM);
	  
	  minWidth = max(minWidth_absPatch[type], minWidth);
	  minHeight = max(minHeight_absPatch[type], minWidth);
	}

	vector<CvRect> rectangles;
	//	if (interestPoints)
	//sampleInterestPoints(samples[i][c], minWidth, minHeight,
	//  _windowSize.width/2, _windowSize.height/2,
	//	     n, rngState, rectangles);
	  //else

	CvMat * mask = NULL;

	if (masks)
	  mask = (*masks)[i];

	sampleLocationsRandomly(minWidth, minHeight,
				_windowSize.width/2, _windowSize.height/2,
				_windowSize.width, _windowSize.height,
				n, rngState, rectangles, mask);


	for (unsigned j = 0; j < rectangles.size(); j++) {
	  CvRect r = rectangles[j];
	  IplImage *t = cvCreateImage(cvSize(r.width, r.height), samples[i][c]->depth, samples[i][c]->nChannels);
	  cvSetImageROI(samples[i][c], r);
	  cvCopyImage(samples[i][c], t);
	  cvResetImageROI(samples[i][c]);
	  
	  // patch is valid over 7x7 pixel region around source location
	  r.x -= 3; r.y -=3; r.width = 7; r.height = 7;
	  
	  svlClipRect(r, _windowSize.width - t->width + 1, 
		      _windowSize.height - t->height + 1);
	  
	  switch (imageTypes[c]) {
	  case SVL_INTENSITY_PATCH:
	    _entries.push_back(new svlIntensityPatchDefinition(t, r, this, c));
	    break;
	  case SVL_DEPTH_PATCH:
	    
	    
	    _entries.push_back(new svlDepthPatchDefinition(t, r, this,  c));

	    assert(_entries.back()->patchType() == SVL_DEPTH_PATCH);

	    numDepthAdded++;
	    break;
	  case SVL_ABS_PATCH:
	    _entries.push_back(new svlAbsPatchDefinition(t, r, this, c, type));
	    break;
	  default:
	    SVL_LOG(SVL_LOG_FATAL, "unrecognized channel type");
	  }
	  
	  cvReleaseImage(&t);
	  
	  // check that this will be a valid dictionary entry
	  if (!_entries.back()->isPatchValid()) {
	    if (_entries.back()->patchType() == SVL_DEPTH_PATCH)
	      {
		numDepthSubtracted++;
	      }
	    _entries.erase(_entries.begin() + _entries.size() - 1);
	  }
	} // patches
      } // images
    } // channel


    if (numDepthAdded)
      {
	if (numDepthSubtracted == numDepthAdded)
	  SVL_LOG(SVL_LOG_WARNING, "All depth patches were rejected!");
      }
  
}

IplImage *svlPatchDictionary::visualizeDictionary() const
{
    int n = (int)ceil(sqrt((double)_entries.size()));
    IplImage *allPatches = cvCreateImage(cvSize(n * _windowSize.width, n * _windowSize.height),
        IPL_DEPTH_8U, 3);
    cvZero(allPatches);

    for (unsigned i = 0; i < _entries.size(); i++) {		
	CvRect r = cvRect((i % n) * _windowSize.width, ((int)(i / n)) * _windowSize.height,
            _windowSize.width, _windowSize.height);
	cvRectangle(allPatches, cvPoint(r.x, r.y), cvPoint(r.x + r.width, r.y + r.height),
            CV_RGB(0, 0, 255), 1);

	cvSetImageROI(allPatches, cvRect(r.x + _entries[i]->_validRect.x, r.y + _entries[i]->_validRect.y,
		_entries[i]->_template->width, _entries[i]->_template->height));
	IplImage *templateCopy = cvCloneImage(_entries[i]->_template);

	if (templateCopy->nChannels != 1) {
	  IplImage *grayTemplateCopy = cvCreateImage(cvGetSize(templateCopy), templateCopy->depth, 1);
	  cvCvtColor(templateCopy, grayTemplateCopy, CV_BGR2GRAY);
	  cvReleaseImage(&templateCopy);
	  templateCopy = grayTemplateCopy;
	}

	scaleToRange(templateCopy, 0.0, 255.0);
        if (templateCopy->depth != IPL_DEPTH_8U) {
            IplImage *tmpImg = cvCreateImage(cvGetSize(templateCopy), IPL_DEPTH_8U, 1);
            cvConvertScale(templateCopy, tmpImg);
            cvReleaseImage(&templateCopy);
            templateCopy = tmpImg;
        }
	for (unsigned k = 0; k < 3; k++) 
	{
	    cvSetImageCOI(allPatches, k + 1);
	    //cvCopyImage(_entries[i]->_template, allPatches);
	    cvCopyImage(templateCopy, allPatches);
	}
	cvReleaseImage(&templateCopy);
	cvSetImageCOI(allPatches, 0);
	cvResetImageROI(allPatches);

	cvRectangle(allPatches, cvPoint(r.x + _entries[i]->_validRect.x, r.y + _entries[i]->_validRect.y),
            cvPoint(r.x + _entries[i]->_validRect.x + _entries[i]->_validRect.width + _entries[i]->_template->width,
                r.y + _entries[i]->_validRect.y + _entries[i]->_validRect.height + _entries[i]->_template->height),
            CV_RGB(255, 0, 0), 1);
    }

    return allPatches;
}

IplImage *svlPatchDictionary::visualizePatch(unsigned index, IplImage *image) const
{
    // create image of the right size (unless one has been passed in)
    if ((image == NULL) || (image->width != _windowSize.width) ||
	(image->height != _windowSize.height) || (image->depth != IPL_DEPTH_8U) ||
	(image->nChannels != 3)) {
	if (image != NULL) {
	    cvReleaseImage(&image);
	}
	image = cvCreateImage(cvSize(_windowSize.width, _windowSize.height),
	    IPL_DEPTH_8U, 3);
    }
    cvZero(image);
    
    // copy template into image
    cvSetImageROI(image, cvRect(_entries[index]->_validRect.x, _entries[index]->_validRect.y,
	    _entries[index]->_template->width, _entries[index]->_template->height));

    IplImage *templateCopy = cvCloneImage(_entries[index]->_template);
    if (templateCopy->nChannels != 1) {
      IplImage *grayTemplateCopy = cvCreateImage(cvGetSize(templateCopy), templateCopy->depth, 1);
      cvCvtColor(templateCopy, grayTemplateCopy, CV_BGR2GRAY);
      cvReleaseImage(&templateCopy);
      templateCopy = grayTemplateCopy;
    }

    scaleToRange(templateCopy, 0.0, 255.0);
    if (templateCopy->depth != IPL_DEPTH_8U) {
        IplImage *tmpImg = cvCreateImage(cvGetSize(templateCopy), IPL_DEPTH_8U, 1);
        cvConvertScale(templateCopy, tmpImg);
        cvReleaseImage(&templateCopy);
        templateCopy = tmpImg;
    }

    for (unsigned k = 0; k < 3; k++) {
	cvSetImageCOI(image, k + 1);
	cvCopyImage(templateCopy, image);
    }
    
    cvReleaseImage(&templateCopy);
    cvSetImageCOI(image, 0);
    cvResetImageROI(image);
    
    // draw box around valid region
    cvRectangle(image, cvPoint(_entries[index]->_validRect.x, _entries[index]->_validRect.y),
	cvPoint(_entries[index]->_validRect.x + _entries[index]->_validRect.width + _entries[index]->_template->width,
	    _entries[index]->_validRect.y + _entries[index]->_validRect.height + _entries[index]->_template->height),
	CV_RGB(255, 0, 0), 1);
    
    // indicate channel index
    for (int j = 0; j <= _entries[index]->_validChannel; j++) {
	CvPoint pt =  cvPoint(_entries[index]->_validRect.x + _entries[index]->_validRect.width + _entries[index]->_template->width,
	    _entries[index]->_validRect.y + _entries[index]->_validRect.height + _entries[index]->_template->height);
	cvLine(image, cvPoint(pt.x - 2 * (j + 1), pt.y - 2), cvPoint(pt.x - 2 * (j + 1), pt.y - 3), CV_RGB(255, 0, 0)); 
    }
    
    return image;
}


ofstream diffs;

/*
TODO-- is this function worth keeping? it does not implement the depth map feature
extraction correctly, so for now I (IG) am commenting it out. The default
svlDetectorFeatureExtractor windowResponse just calls extract on a single location
with the sparse flag asserted

void svlPatchDictionary::windowResponse(const vector<IplImage *> & images,
					vector<double> & output,
					int outputOffset,
					const vector<bool> *usedFeatures) const
{
  if (output.size() < outputOffset + _entries.size())
    output.resize(outputOffset + _entries.size());

    assert(!images.empty());
    for (unsigned c = 0; c < images.size(); c++) {
        assert(images[c] != NULL);
        if ((images[c]->width != _windowSize.width) ||
            (images[c]->height != _windowSize.height)) {
            // TO DO: resize
            SVL_LOG(SVL_LOG_FATAL, "invalid image size in patchResponse");
        }
    }



    for (unsigned i = 0; i < _entries.size(); i++) 
	{
	 

//TODO-- re-enable this check but put it somewhere where it's not in usch a tight loop	
	//  if ((unsigned)_entries[i]->_validChannel >= images.size())
	//    {
	//      int count = _entries[i]->_validChannel;
	//      cout << "Patch dictionary wants to access channel ";
	//      cout << count;
	//      cout <<" but there are only "<<images.size()<<" channels."<<endl;
	//      assert(false);
	//      } //todo-- reenable this check but put it somewhere where it's not in such a tight loop

	  // IplImage *response = _entries[i]->responseImage(images[_entries[i]->_validChannel]);

	  //fv[i] = _entries[i]->patchValue(response, cvPoint(0, 0));
	  //cvReleaseImage(&response);

	  if (usedFeatures && !usedFeatures->at(i))
	    output[outputOffset+i] = 0;
	  else 
	    output[outputOffset+i]  =  _entries[i]->trainingResponseValue(images[_entries[i]->_validChannel]);
	}

}
*/



void svlPatchDictionary::normalizeDepthPatches(IplImage *image, const vector<CvPoint> &windows) const {
   _normalizedDepthPatches.clear();


    //Load depth map image
    SVL_ASSERT_MSG(image->nChannels == 1, "in svlPatchDictionary::imageResponse, the image at the depth channel is not grayscale");

    // Populate normalized depth patches
#ifdef USE_THREADS
    _normalizedDepthPatches.resize(windows.size());
    svlThreadPool threadPool(TOTAL_THREADS);
    const int hThreadNormalization = svlCodeProfiler::getHandle("svlPatchDictionary::Normalize depth patches synchronously");
    svlCodeProfiler::tic(hThreadNormalization);
    threadPool.start();
#endif

    for (unsigned j = 0; j < windows.size(); j++) 
    {
      IplImage *patch = cvCreateImage(_windowSize, image->depth, 1);
      cvSetImageROI(image, cvRect(windows[j].x, windows[j].y, _windowSize.width, _windowSize.height));
      cvCopyImage(image, patch);
      cvResetImageROI(image);
            
      // Normalize the entire patch by subtracting the median
#ifdef USE_THREADS
      threadPool.addJob(normalizePatch, new NormalizePatchArgs(this, patch, j));
#else
      vector<float> depths(patch->width * patch->height);            
      int i = 0;
      for(int y = 0; y < patch->height; y++) {
          for(int x = 0; x < patch->width; x++) {
              depths[i++] = CV_IMAGE_ELEM(patch, float, y, x);
          }
      }
            
      sort(depths.begin(),depths.end());

      float median = depths[depths.size()/2];
            
      for(int y = 0; y < patch->height; y++) {
          for(int x = 0; x < patch->width; x++) {
              CV_IMAGE_ELEM(patch, float, y, x) -= median;
          }
      }

      depths.clear();

      svlPatchDictionary::_normalizedDepthPatches.push_back(patch);
#endif // USE_THREADS

    } // windows loop

#ifdef USE_THREADS
    // wait for finish
    threadPool.finish();
    svlCodeProfiler::toc(hThreadNormalization);
#endif

}

// todo: also potentially make threaded
void svlPatchDictionary::computeAbsResponses(IplImage *image) const {
  // create the image responses for each entry type
  _responses.resize(SVL_ABS_NUM, NULL);
  _response_origin.resize(SVL_ABS_NUM, -1);
  
  for (unsigned i = 0; i < _entries.size(); i++) {
    if (_entries[i]->patchType() != SVL_ABS_PATCH) continue;
    
    svlAbsPatchDefinition *patch = (svlAbsPatchDefinition *)((svlPatchDefinition *)(_entries[i])); 
    svlAbsPatchType t = patch->getAbsType();
    if (_responses[t] != NULL) continue;
    
    _responses[t] = patch->maskResponseImages(image);
    _response_origin[t] = i;
  }
}

void svlPatchDictionary::clearAbsResponses()  const {
  for (unsigned i = 0; i < _responses.size(); i++) {
      if (_responses[i] != NULL) {
	int j = _response_origin[i];
	svlAbsPatchDefinition *orig_patch = (svlAbsPatchDefinition *)((svlPatchDefinition *)(_entries[j]));
	orig_patch->deallocateResponseImages(_responses[i]);
      }
    }

  _responses.clear();
  _response_origin.clear();
}


void svlPatchDictionary::extract(const vector<CvPoint> &windows, 
				 const vector<svlDataFrame*> &frames,
				 vector< vector<double> > &output,
				 bool sparse,
				 int outputOffset) const 
{

  //Convert list of frames to list of IplImages, using NULL for frames that are not images
  vector<IplImage *> images(frames.size(), NULL);

  for (unsigned i = 0; i < frames.size(); i++)
    {
      assert(frames[i]);
      svlImageFrame * temp = dynamic_cast<svlImageFrame *>(frames[i]);
      if (temp)
	images[i] = temp->image;
    }



  // if input is sparse, then integral images are not used
  bool useIntegralImages = _bUseIntImg && (!sparse);

  // allocate memory for return vectors
  if (output.size() != windows.size())
    output.resize(windows.size());

  unsigned neededSize = outputOffset + _entries.size();

  for (unsigned i = 0; i < windows.size(); i++)  {
    if (output[i].size() < neededSize)
      output[i].resize(neededSize);
  }

  int numChannels = (int)images.size();

  svlDepthPatchDefinition::_windowSize = _windowSize;

  //Check if any channels are depth or abs
  vector<bool> isDepthChannel(frames.size(),false);
  vector<bool> isAbsChannel(frames.size(), false);  

  for (unsigned i = 0; i < _entries.size(); i++)
    {
      if (_entries[i]->patchType() == SVL_DEPTH_PATCH)
	isDepthChannel[ _entries[i]->validChannel() ] = true;

      if (_entries[i]->patchType() == SVL_ABS_PATCH)
	isAbsChannel[ _entries[i]->validChannel() ] = true;
    }

  for (unsigned i = 0; i < frames.size(); i++)
    {
      if (isDepthChannel[i])
	normalizeDepthPatches(images[i], windows);

      if (isAbsChannel[i])
	computeAbsResponses(images[i]);
    }

  IplImage **imageSum = NULL;
  IplImage **imageSumSq = NULL;

  if (useIntegralImages) {
    // pre-compute integral images for use by patch responses
    imageSum = new IplImage* [numChannels];
    imageSumSq = new IplImage* [numChannels];

    const int hIntegralImages = svlCodeProfiler::getHandle("svlPatchDictionary::Precompute integral images");
    svlCodeProfiler::tic(hIntegralImages);

    for (int c = 0; c < numChannels; c++) {
	imageSum[c] = NULL;
	imageSumSq[c] = NULL;
   
	if (images[c] == NULL || isDepthChannel[c] || isAbsChannel[c])
	    continue; 
	
	if (images[c]->depth == IPL_DEPTH_8U) {
	    imageSum[c] = cvCreateImage(cvSize(images[c]->width + 1, images[c]->height + 1),
					IPL_DEPTH_32S, images[c]->nChannels);
	    imageSumSq[c] = cvCreateImage(cvSize(images[c]->width + 1, images[c]->height + 1),
					  IPL_DEPTH_64F, images[c]->nChannels);
	    cvIntegral(images[c], imageSum[c], imageSumSq[c]);
	}
    }
    
    svlCodeProfiler::toc(hIntegralImages);
  }

#ifdef USE_THREADS
  int handle = svlCodeProfiler::getHandle("svlPatchDefinition::patchValue (threaded)");
  svlCodeProfiler::tic(handle);
  svlThreadPool vThreadPool(TOTAL_THREADS);
  vThreadPool.start();
#endif

  for (unsigned i = 0; i < _entries.size(); i++) 
  {
    IplImage *img = images[_entries[i]->_validChannel];
    IplImage *imgSum = NULL;
    IplImage *imgSumSq = NULL;
 
    if (useIntegralImages) {
      imgSum = imageSum[_entries[i]->_validChannel];
      imgSumSq = imageSumSq[_entries[i]->_validChannel];
    }

    CalcFVArgs *args = new CalcFVArgs(img, imgSum, imgSumSq, &_responses, sparse,
				      &_entries, i, &windows, &output, outputOffset);

#ifdef USE_THREADS
    vThreadPool.addJob(calcFV, args);
#else
    int handle = svlCodeProfiler::getHandle("svlPatchDefinition::patchValue");
    svlCodeProfiler::tic(handle);
    calcFV(args, 0);
    svlCodeProfiler::toc(handle);
#endif

  }
#ifdef USE_THREADS
  vThreadPool.finish();
  svlCodeProfiler::toc(handle);
#endif

  // Free all memory from the normalized patches

  for (unsigned j = 0; j < svlPatchDictionary::_normalizedDepthPatches.size();j++) 
  {
    cvReleaseImage(&svlPatchDictionary::_normalizedDepthPatches[j]);
  }
  
  _normalizedDepthPatches.clear();

  // Free all memory from response images

  clearAbsResponses();

  // Free all memory from integral images

  if (useIntegralImages) {
    // release integral images
    for (int i = 0; i < numChannels; i++) 
      {
        if (imageSumSq[i] != NULL)
	  cvReleaseImage(&imageSumSq[i]);
        if (imageSum[i] != NULL)
	  cvReleaseImage(&imageSum[i]);
      }
    
    delete[] imageSumSq;
    delete[] imageSum;
  }
}


svlDetectorFeatureExtractor * svlPatchDictionary::clone() const
{
  return dynamic_cast<svlDetectorFeatureExtractor *>(new svlPatchDictionary(*this));
}


svlDetectorFeatureExtractor* svlPatchDictionary::getPrunedExtractor
      (const vector<bool> &featureUsedBits) const
{
  svlPatchDictionary * temp = new svlPatchDictionary(*this);
  assert(temp);
  temp->filterEntries(featureUsedBits);
  return dynamic_cast<svlDetectorFeatureExtractor *>(temp);
}


string svlPatchDictionary::summary() const
{
  unsigned numIntensity = 0;
  unsigned numDepth = 0;
  unsigned numAbs = 0;

  for (unsigned i = 0; i < _entries.size(); i++)
    {
      svlPatchDefinitionType t = _entries[i]->patchType();
      switch (t)
	{
	case SVL_DEPTH_PATCH:
	  numDepth++;
	  break;
	case SVL_INTENSITY_PATCH:
	  numIntensity++;
	  break;
	case SVL_ABS_PATCH:
	  numAbs++;
	  break;
	default:
	  SVL_LOG(SVL_LOG_FATAL, "svlPatchDictionary::summarize encountered unknown patchtype");
	}
    }

  stringstream s;
  s << "[";

  bool first = true;

  if (numIntensity)
    {
      s << numIntensity << " intensity patches";
      first = false;
    }

  if (numDepth)
    {
      if (first)
	first = false;
      else
	s << ", ";

      s << numDepth << " depth patches";
    }

  if (numAbs)
    {
      if (first)
	first = false;
      else
	s << ", ";

      s << numAbs << "abs patches";
    }

  s << "]";


  return s.str();
}
