/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlPatchUtils.cpp
** AUTHOR(S):   Ian Goodfellow <ia3n@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**      based on applications by Stephen Gould and Sid Batra
**             
** DESCRIPTION:
**   Commonly re-used subroutines for processing command line arguments related
**   to patch types, mapping from channel types to patch types, or reading
**   and writing image files in the patch based object detector pipeline
**
*****************************************************************************/
#include "svlPatchUtils.h"

static map<string, svlChannelType> svlSetupTextToChannelConverter();
static map<string, svlColorType> svlSetupTextToColorConverter();


svlPatchUtil::svlPatchUtil(unsigned flags) : 
  _baseExt(".jpg"), //_baseExt("")
  _cliBaseExt(false),
  _defaultEdge(false),
  _cliMask(false),
  _mask(false),
  _maskExt(".occ"),
  _depthMapWidth(-1),
  _depthMapHeight(-1),
  _width(-1),
  _height(-1),
    _depthMapExt(".reconstructed.z.txt"),
    _imageExt(".jpg"),
  _cliDepthChannel(false),
    _hasDepthChannel(false),
    _usedImageExt(false),
  _usedDepthMapExt(false),
  _usedDeprecatedCh(false),
  _usedNewCh(false)
{

  // Load the maps to parse the text
  _textToChannelType = svlSetupTextToChannelConverter();
  _textToColorType = svlSetupTextToColorConverter();

  //Load the mapping from channel type to patch type
  _channelTypeToPatchType[SVL_CHANNEL_INTENSITY] = SVL_INTENSITY_PATCH;
  _channelTypeToPatchType[SVL_CHANNEL_EDGE] = SVL_INTENSITY_PATCH;
  _channelTypeToPatchType[SVL_CHANNEL_DEPTH] = SVL_DEPTH_PATCH;
  _channelTypeToPatchType[SVL_CHANNEL_ABS] = SVL_ABS_PATCH;
  
    if (flags & SVL_PATCH_DEFAULT_EDGE)
	_defaultEdge = true;
     

    if (flags & SVL_PATCH_CLI_DEPTHCHANNEL)
      _cliDepthChannel = true;

    if (flags & SVL_PATCH_CLI_BASE_EXT)
      _cliBaseExt = true;

    if (flags & SVL_PATCH_CLI_MASK)
      _cliMask = true;
}

svlPatchUtil::svlPatchUtil(const svlPatchUtil & o) :
  _channelTypes(o._channelTypes), _channelPatchTypes(o._channelPatchTypes), _colorTypes(o._colorTypes),
  _channelExts(o._channelExts), 
  _textToChannelType(o._textToChannelType), _textToColorType(o._textToColorType), 
  _channelTypeToPatchType(o._channelTypeToPatchType),
  _baseExt(o._baseExt), _cliBaseExt(o._cliBaseExt), _defaultEdge(o._defaultEdge),
  _cliMask(o._cliMask), _mask(o._mask), _maskExt(o._maskExt),
  _dir(o._dir), _eTree(o._eTree), _files(o._files),
  _depthMapWidth(o._depthMapWidth), _depthMapHeight(o._depthMapHeight), _width(o._width), _height(o._height),
  _softEdgeCalculator(o._softEdgeCalculator), 
  _depthMapExt(o._depthMapExt), _imageExt(o._imageExt),
  _cliDepthChannel(o._cliDepthChannel), _hasDepthChannel(o._hasDepthChannel), 
  _usedImageExt(o._usedImageExt), _usedDepthMapExt(o._usedDepthMapExt), 
  _usedDeprecatedCh(o._usedDeprecatedCh), _usedNewCh(o._usedNewCh)
{
}


bool svlPatchUtil::parseCommandLineItem(int & argc, char ** & args)
{
  if (_cliBaseExt && !strcmp(*args, "-baseExt"))
    {
      _baseExt = * (++args);
      argc--;
    }
  else if (_cliMask && !strcmp(*args, "-mask"))
    {
      _mask = true;
      _maskExt = * (++args);
      argc--;
    }
  else if (! strcmp(*args, "-ignore")) //use this to tell eTree not to load files ending in some extension. TODO-- make patch dictionary smart enough to ignore arbitrary channels, right now you can only ignore channels at the end or the patch dictionary will come out with the wrong validChannel indices
    {
      _eTree.ignore(*(++args)); argc--;
    }
  else if (!strcmp(*args, "-imageExt"))
    {
      SVL_LOG(SVL_LOG_WARNING, "-imageExt is deprecated, use -ch INTENSITY <ext> instead");
      _usedImageExt = true;
      _imageExt = string(*(++args)); argc--;
    } 
  else if (!strcmp(*args, "-depthMapExt"))
    {
      SVL_LOG(SVL_LOG_WARNING, "-depthMapExt is deprecated, use -ch DEPTH <ext> instead");
      _depthMapExt = *(++args); 
      argc--;
      _usedDepthMapExt = true;
    }
  else if (!strcmp(*args, "-dmsize")) 
    {
      _depthMapWidth = atoi(*(++args)); argc--;
      _depthMapHeight = atoi(*(++args)); argc--;
    }
  else if (_cliDepthChannel && !strcmp(*args, "-depthChannel"))
    {
      SVL_LOG(SVL_LOG_WARNING, "-depthChannel is deprecated. use -ch DEPTH <ext> instead");
      _hasDepthChannel = true;
    }
  else if (!strcmp(*args, "-ch"))
      {
	//Advance to the first argument to the flag-- should be a type if using the new system or a number if using the deprecated system
	++args;
	argc--;

	if (argc == 0)
	  SVL_LOG(SVL_LOG_FATAL, "No arguments to -ch flag");

	//See if we can parse the next argument as a type
	if (_textToChannelType.find(*args) != _textToChannelType.end())
	  {
	    //We parsed the argument as a type, so use the new system

	    //Record that we used this system
	    _usedNewCh = true;

	    //Since we're using the new system, make sure the deprecated system hasn't been used
	    if (_usedDeprecatedCh)
	      SVL_LOG(SVL_LOG_FATAL, "Attempted to use deprecated -ch format with other non-deprecated channel arguments. This mixture of syntax is not supported");

	    //Get the channel type
	    svlChannelType chType = _textToChannelType[*args];
	    svlColorType crType = _textToColorType[*args];

	    //Read to the next argument
	    args++;
	    argc--;

	    if (argc == 0)
	      SVL_LOG(SVL_LOG_FATAL, "Only one argument to -ch flag, two required");

	    //Save the type
	    _channelTypes.push_back(chType);
	    _colorTypes.push_back(crType);
	    
	    //Save the patch type
	    _channelPatchTypes.push_back( _channelTypeToPatchType[chType] );

	    //Save the extension for this channel (the next argument)
	    _channelExts.push_back(*args);
	  }
	else
	  {
	    //In this case, we could not parse the next argument as a type
	    //This could be a syntax error or the deprecated usage of -ch
	    
	    int numChannels = atoi(*args);
	    
	    //Check to see if this is a syntax error
	    if (numChannels == 0)
	      SVL_LOG(SVL_LOG_FATAL, "Error while processing -ch flag. Either first arg (\""<<(*args)<<"\")is syntax error, or request for 0 channels in deprecated usage of -ch");

	    //If we got to here, this is the deprecated usage

	    SVL_LOG(SVL_LOG_WARNING, "using deprecated version of -ch");

	    //Record that this deprecated argument has been used
	    _usedDeprecatedCh = true;
	 
	    //Make sure no other channels have been created by other means
	    if (_channelTypes.size() != 0)
	      SVL_LOG(SVL_LOG_FATAL, "Attempted to use deprecated -ch format with other non-deprecated channel arguments. This mixture of syntax is not supported");
	    

	    //Read all of the channel types
	    for (int i = 0; i < numChannels; i++)
	      {
		args++;
		argc--;

		//Check that we haven't read past the end of input
		if (argc == 0)
		  SVL_LOG(SVL_LOG_FATAL, "" << numChannels << " channels requested with deprecated version of -ch, but only "<<i<<" channel types specified");

		//Check that the argument is a valid type
		if (_textToChannelType.find(*args) == _textToChannelType.end())
		  SVL_LOG(SVL_LOG_FATAL, "Invalid channel type \""<<*args<<"\" requested in call to deprecated version of -ch")
		else
		  {
		    //Get the channel type
		    svlChannelType chType = _textToChannelType[*args];
		    svlColorType crType = _textToColorType[*args];
		    //Get the patch type
		    svlPatchDefinitionType pType = _channelTypeToPatchType[chType];

		    //Store the channel type and patch type (channel extension will have to be back-patched later because this is the deprecated syntax)
		    _channelTypes.push_back(chType);
		    _colorTypes.push_back(crType);
		    _channelPatchTypes.push_back(pType);
		  }
	      }//closes loop over all the channel type arguments
	  } //closes processing of deprecated version of ch arg or syntax error following -ch arg
      } //closes processing of -ch arg
  else
    return false;

  return true;
}

void svlPatchUtil::finishCommandLine()
{
  if (_usedNewCh)
    {
      assert(!_usedDeprecatedCh); //We already have checks against this so if both are true something is wrong with the parser
     

      //We should check the usage though
      if (_usedImageExt)
	SVL_LOG(SVL_LOG_FATAL, "Attempted to use new -ch argument with deprecated -imageExt argument. This hybrid usage is not supported");
      if (_usedDepthMapExt)
	SVL_LOG(SVL_LOG_FATAL, "Attempted to use new -ch argument with deprecated -depthMapExt argument. This hybrid usage is not supported");
      if (_hasDepthChannel)
	SVL_LOG(SVL_LOG_FATAL, "Attempted to use new -ch argument with deprecated -depthChannel argument. This hybrid usage is not supported.");
    }

  if (_channelTypes.size())
    {
      //The channels have been populated already, using some form of -ch

      //If the user used the deprecated version of -ch, then the extensions won't be populated yet
      if (!_channelExts.size())
	{
	  //If this happens it should only be because of use of the deprecated version of -ch
	  assert(_usedDeprecatedCh);

	  //Fill in the extension based on the type for all channels
	  for (unsigned i = 0; i < _channelTypes.size(); i++)
	    {
	      switch (_channelTypes[i])
		{
		case SVL_CHANNEL_DEPTH:
		  _channelExts.push_back(_depthMapExt);
		  break;
		default:
		  _channelExts.push_back(_imageExt);
		  break;
		} //closes the switch over channel types
	    }//closes the loop over channels
	}//closes handling of unpopulated channelExts
      else
	assert( !_usedDeprecatedCh); //If the extensions were populated, then we must not have used the deprecated -ch
    }//closes case where channels are already populated
  else
    {
      //The channels have not been populated yet, so we must fill in the defaults

      //Always assume there is an INTENSITY channel.
      _channelTypes.push_back(SVL_CHANNEL_INTENSITY);
      _colorTypes.push_back(SVL_COLOR_GRAY);
      _channelPatchTypes.push_back(SVL_INTENSITY_PATCH);
      //Default its extension to baseExt, but for compatibility with old scripts, respect the deprecated imageExt flag if it was called
      if (_usedImageExt)
	_channelExts.push_back(_imageExt);
      else
	_channelExts.push_back(_baseExt);

      //If configured to assume there is an edge patch, add an edge patch
      if (_defaultEdge)
	{
	  _channelTypes.push_back(SVL_CHANNEL_EDGE);
	  _colorTypes.push_back(SVL_COLOR_GRAY);
	  _channelPatchTypes.push_back(SVL_INTENSITY_PATCH);
	  if (_usedImageExt)
	    _channelExts.push_back(_imageExt);
	  else
	    _channelExts.push_back(_baseExt);
	}

      //For compatibility with old scripts, respect the deprecated depthChannel flag if it was called
      if (_hasDepthChannel)
	{
	  _channelTypes.push_back(SVL_CHANNEL_DEPTH);
	  _colorTypes.push_back(SVL_COLOR_UNDEFINED);
	  _channelPatchTypes.push_back(SVL_DEPTH_PATCH);
	  _channelExts.push_back(_depthMapExt);
	}
    }

  //Patch any uses of "-" with baseExt
  for (unsigned i = 0; i < _channelExts.size(); i++) {
    if (_channelExts[i] == "-")
      _channelExts[i] = _baseExt;
  }
  

  //At this point all of our channel definition vectors should be complete
  //Check that their sizes match
  assert(_channelTypes.size() == _colorTypes.size());
  assert(_channelTypes.size() == _channelPatchTypes.size());
  assert(_channelPatchTypes.size() == _channelExts.size());
  //Check that they are non-empty
  assert(_channelTypes.size() != 0);

  //Prepare for loading files later on
  for (unsigned i = 0; i < _channelExts.size(); i++)
    {
      _files.push_back(vector<string> ());
      _eTree.addChannel(i, _channelExts[i]);
    }

  if (_mask)
    {
      _files.push_back(vector<string> ());
      _eTree.addChannel(_channelExts.size(), _maskExt);
    }

}

void svlPatchUtil::readDir(const char * dir)
{
  _dir = dir;

  _eTree.readDirectory(dir);

  for (unsigned i = 0; i < _files.size(); i++)
    {
      _files[i].clear();
      _eTree.getFilesForChannel(i, _files[i]);
    }


  //Check that the files all correspond to each other
  if (_files.size() > 1)
    {
      for (unsigned i = 1; i < _files.size(); i++)
	{
	  string extension;

	  if (i < _channelExts.size())
	    extension = _channelExts[i];
	  else
	    extension = _maskExt;

	  if (_files[i].size() != _files[0].size())
	    SVL_LOG(SVL_LOG_FATAL, "Channel "<<i<<" and channel 0 do not have the same number of frames: channel "<<i<<" has "<<_files[i].size()<<" while channel 0 has "<<_files[0].size()<<". (Aborting after first difference)");
	  
	  for (unsigned j = 0; j < _files[0].size(); j++)
	    {
	      assert(!strcmp(stripExtension(_files[0][j], _channelExts[0]).c_str(), stripExtension(_files[i][j], extension).c_str()));
	    }
	}
    }
}

void svlPatchUtil::getAllFrames(vector<vector<IplImage *> > & output, unsigned maxToRead)
{
  output.clear();

  bool newFrameNeeded = true;
  for (unsigned int i = 0; i < _files[0].size() && i < maxToRead; i++)
    {
      if (newFrameNeeded)
	output.push_back(vector<IplImage *>() );

      newFrameNeeded = getFrame(i, output.back());
    }
}

void svlPatchUtil::getAllFrames(vector<svlMaskedFrame> & output) 
{
  output.clear();

  bool newFrameNeeded = true;
  for (unsigned int i = 0; i < _files[0].size(); i++)
    {
      if (newFrameNeeded)
	output.push_back(svlMaskedFrame() );

      newFrameNeeded = getFrame(i, output.back());
    }
}

bool svlPatchUtil::getFrame(unsigned idx, vector<IplImage *> & output) 
{
  SVL_LOG(SVL_LOG_MESSAGE, "Getting frame " << stripExtension(_files[0][idx], _channelExts[0]));

  vector<string> filenames;

  for (unsigned ch = 0; ch < _channelTypes.size(); ch++)
    filenames.push_back(  _dir + "/" + _files[ch][idx] );

  return getFrameFromFiles(filenames, output);
}

bool svlPatchUtil::getFrame(unsigned idx, svlMaskedFrame & output) 
{
if (_mask)
{
  SVL_LOG(SVL_LOG_MESSAGE, "Getting masked frame " << stripExtension(_files[0][idx], _channelExts[0]));
}else  
SVL_LOG(SVL_LOG_MESSAGE, "Getting frame " << stripExtension(_files[0][idx], _channelExts[0]));

  output.channels.clear();
  
  if (output.mask)
    cvReleaseMat(&output.mask);

  vector<string> filenames;

  for (unsigned ch = 0; ch < _channelTypes.size(); ch++)
    filenames.push_back(  _dir + "/" + _files[ch][idx] );

  bool success1 = getFrameFromFiles(filenames, output.channels);

  if (success1)
    {
      if (_mask)
	{
	  assert(output.channels.size());
	  output.mask = cvCreateMat(output.channels[0]->height,
				     output.channels[0]->width,
				     CV_32FC1);
	  assert(output.mask);

	  bool success2 = readMatrix(output.mask, (_dir + "/" + _files[numChannels()][idx]).c_str() );

	  return success2;
	}
      else
	{
	  output.mask = NULL;
	}

      return true;
    }
  else
    return false;

  //unreached
  assert(false);
  return false;
}


bool svlPatchUtil::getFrameFromBasename(const string & baseName, vector<IplImage *> & output) 
{
  string strippedName = baseName.substr(0,baseName.size()-_baseExt.size());

  vector<string> filenames;

  for (unsigned ch = 0; ch < _channelExts.size(); ch++)
    filenames.push_back(strippedName + _channelExts[ch]);

  return getFrameFromFiles(filenames, output);
}

bool svlPatchUtil::getFrameFromFiles(const vector<string> & files, vector<IplImage *> & output) 
{

  if (files.size() != _channelTypes.size())
    return false;

  for (unsigned ch = 0; ch < _channelTypes.size(); ch++)
    {
      //TODO-- this implementation doesn't attempt to avoid loading the same file multiple times. Could restore that feature
      //note-- reloading the same file doesn't seem to incur much of a performance penalty in my experiments, in fact,
      //my new code came out faster than the original in all three trials that I ran. I assume this is because the file is reloaded soon after
      //the initial load and is still in the disk buffer. The short story is that I don't think avoiding reloading the same file
      //is a high priority. -IG

      // Certainly makes the code a heck of a lot cleaner and happier. -- OR

      const string & filename = files[ch];

      IplImage * img = loadFrame(filename, ch);

      if (img == NULL)
	{
	  SVL_LOG(SVL_LOG_MESSAGE, "...failed to load " << filename);

	  for (unsigned ch2 = 0; ch2 < ch; ch2++)
	    cvReleaseImage(&output[ch2]);

	  output.clear();

	  return false;
	}
      else
	{
	  if ((img->width != (int) _width) || (img->height != (int) _height))
	    {
	      if (_height != -1)
		{
		  //cout << "Calling resizeInPlace "<<_height<<" "<<_width<<endl;
		  resizeInPlace(&img, _height, _width);
		}
	    }

	  if (_channelTypes[ch] == SVL_CHANNEL_EDGE) {
	    IplImage *temp = cvCloneImage(_softEdgeCalculator.processImage(img));
	    cvReleaseImage(&img);
	    img = temp;
	  }
	  
	  output.push_back(img);	  
	}    
    }


  return true;
}

IplImage *svlPatchUtil::loadFrame(const string & path, unsigned channel) const {
  IplImage *img = NULL;

  svlColorType cType = _colorTypes[channel];
  switch (cType)
    {
    case SVL_COLOR_GRAY:
      img = cvLoadImage(path.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
      break;
    case SVL_COLOR_UNDEFINED:
      //      cout << "dmsize: "<<_depthMapWidth<< " "<<_depthMapHeight<<endl;
      img = readMatrixAsIplImage(path.c_str(), _depthMapWidth, _depthMapHeight);
      break;
    default:
      cout << "Loading unchanged" << endl;
      img = cvLoadImage(path.c_str(), CV_LOAD_IMAGE_UNCHANGED);
        // do the necessary color conversion for non-grayscale images
      svlChangeColorModel(img, cType);
    }
  return img;
}


void svlPatchUtil::setDepthMapSize(int width, int height)
{
  _depthMapWidth = width;
  _depthMapHeight = height;
}

void svlPatchUtil::setResizeSize(int width, int height)
{

  _width = width;
  _height = height;

  if (_depthMapWidth == -1)
    setDepthMapSize(width, height);
}

unsigned svlPatchUtil::numFrames() const
{
  if (_files.size())
    return (unsigned)_files[0].size();
  return 0;
}


string svlPatchUtil::stripExtension(const string & str, const string & ext)
{
  assert(str.size() >= ext.size());

  return str.substr(0, str.size()-ext.size());
}



// messy helper functions

static map<string, svlChannelType> svlSetupTextToChannelConverter() {
  map<string, svlChannelType> converter;

  //Load the mapping from command line text to channel types
  converter["INTENSITY"] = SVL_CHANNEL_INTENSITY;  
  converter["COLOR"] = SVL_CHANNEL_INTENSITY;
  
  converter["EDGE"] = SVL_CHANNEL_EDGE;
  converter["EDGE_RGB"] = SVL_CHANNEL_EDGE;
  converter["EDGE_BG"] = SVL_CHANNEL_EDGE;
  converter["EDGE_BR"] = SVL_CHANNEL_EDGE;
  converter["EDGE_GR"] = SVL_CHANNEL_EDGE;
  converter["EDGE_HSV"] = SVL_CHANNEL_EDGE;
  converter["EDGE_HS"] = SVL_CHANNEL_EDGE;
  converter["EDGE_HV"] = SVL_CHANNEL_EDGE;
  converter["EDGE_YCrCb"] = SVL_CHANNEL_EDGE;
  converter["EDGE_CrCb"] = SVL_CHANNEL_EDGE;
  
  converter["DEPTH"] = SVL_CHANNEL_DEPTH;

  converter["ABSG"] = SVL_CHANNEL_ABS;
  converter["ABS"] = SVL_CHANNEL_ABS;
  converter["ABS_BG"] = SVL_CHANNEL_ABS;
  converter["ABS_BR"] = SVL_CHANNEL_ABS;
  converter["ABS_GR"] = SVL_CHANNEL_ABS;
  converter["ABS_HSV"] = SVL_CHANNEL_ABS;
  converter["ABS_HS"] = SVL_CHANNEL_ABS;
  converter["ABS_HV"] = SVL_CHANNEL_ABS;
  converter["ABS_YCrCb"] = SVL_CHANNEL_ABS;

  return converter;
}

static map<string, svlColorType> svlSetupTextToColorConverter() {
  map<string, svlColorType> converter;

  //Load the mapping from command line text to channel types
  converter["DEPTH"] = SVL_COLOR_UNDEFINED;

  converter["INTENSITY"] = SVL_COLOR_GRAY;  
  converter["EDGE"] = SVL_COLOR_GRAY;
  converter["ABSG"] = SVL_COLOR_GRAY;

  converter["EDGE_RGB"] = SVL_COLOR_BGR;
  converter["ABS"] = SVL_COLOR_BGR;

  converter["EDGE_BG"] = SVL_COLOR_BG;
  converter["ABS_BG"] = SVL_COLOR_BG;

  converter["EDGE_BR"] = SVL_COLOR_BR;
  converter["ABS_BR"] = SVL_COLOR_BR;

  converter["EDGE_GR"] = SVL_COLOR_GR;
  converter["ABS_GR"] = SVL_COLOR_GR;

  converter["EDGE_HSV"] = SVL_COLOR_HSV;
  converter["ABS_HSV"] = SVL_COLOR_HSV;

  converter["EDGE_HS"] = SVL_COLOR_HS;
  converter["ABS_HS"] = SVL_COLOR_HS;

  converter["EDGE_HV"] = SVL_COLOR_HV;
  converter["ABS_HV"] = SVL_COLOR_HV;

  converter["EDGE_YCrCb"] = SVL_COLOR_YCrCb;
  converter["ABS_YCrCb"] = SVL_COLOR_YCrCb;

  converter["EDGE_CrCb"] = SVL_COLOR_CrCb;

  converter["COLOR"] = SVL_COLOR_H;
  
  return converter;
}



